import docutils


class SectNumFolder(docutils.nodes.SparseNodeVisitor):
    def __init__(self, document):
        docutils.nodes.SparseNodeVisitor.__init__(self, document)
        self.sectnums = {}

    def visit_generated(self, node):
        for i in node.parent.parent['ids']:
            self.sectnums[i] = node.parent.astext().replace(u'\xa0\xa0\xa0', ' ')

    def unknown_visit(self, node):
        pass


class SectRefExpander(docutils.nodes.SparseNodeVisitor):
    def __init__(self, document, sectnums):
        docutils.nodes.SparseNodeVisitor.__init__(self, document)
        self.sectnums = sectnums

    def visit_reference(self, node):
        if node.get('refid', None) in self.sectnums:
            node.children = [
                docutils.nodes.Text('%s ' % self.sectnums[node.get('refid')])
            ]

    def unknown_visit(self, node):
        pass



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

from StringIO import StringIO

from docutils import writers

from . import createpdf


class PdfWriter(writers.Writer):
    def __init__(self, builder):
        writers.Writer.__init__(self)
        self.builder = builder
        self.output = u''

    supported = 'pdf'
    """Formats this writer supports."""

    config_section = 'pdf writer'
    config_section_dependencies = 'writers'
    """Final translated form of `document`."""

    def translate(self):
        sio = StringIO('')
        createpdf.RstToPdf(sphinx=True).createPdf(
            doctree=self.document, output=sio, compressed=False
        )
        self.output = str(sio.getvalue(), 'utf-8', 'ignore')

    def supports(self, format):
        """This writer supports all format-specific elements."""
        return 1



# -*- coding: utf-8 -*-

from reportlab.platypus.flowables import _listWrapOn, _FUZZ, Flowable


class Sinker(Flowable):
    """A flowable that always takes the rest of the frame.
    It then draws its contents (a list of sub-flowables)
    at the bottom of that space"""

    def __init__(self, content):
        self.content = content

    def wrap(self, aW, aH):
        self.width, self.height = _listWrapOn(self.content, aW, None)
        return self.width, aH

    def draw(self):
        canv = self.canv
        canv.saveState()
        x = canv._x
        y = canv._y
        y += self.height
        aW = self.width
        for c in self.content:
            w, h = c.wrapOn(canv, aW, 0xFFFFFFF)
            if (w < _FUZZ or h < _FUZZ) and not getattr(c, '_ZEROSIZE', None):
                continue
            y -= h
            canv.saveState()
            c.drawOn(canv, x, y, _sW=aW - w)
            canv.restoreState()
        canv.restoreState()



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms
'''
Creates a rst2pdf stylesheet for each pygments style.
'''

import os

from dumpstyle import dumps
from pygments.token import STANDARD_TYPES
from pygments import styles as pstyles


# First get a list of all possible classes
classnames = set()
for name in list(pstyles.get_all_styles()):
    css = os.popen('pygmentize -S %s -f html' % name, 'r').read()
    for line in css.splitlines():
        line = line.strip()
        sname = "pygments-" + line.split(' ')[0][1:]
        classnames.add(sname)


def css2rl(css):
    dstyles = {}
    # First create a dumb stylesheet
    for key in STANDARD_TYPES:
        dstyles["pygments-" + STANDARD_TYPES[key]] = {'parent': 'code'}
    seenclassnames = set()
    styles = []
    for line in css.splitlines():
        line = line.strip()
        sname = "pygments-" + line.split(' ')[0][1:]
        seenclassnames.add(sname)
        style = dstyles.get(sname, {'parent': 'code'})
        options = line.split('{')[1].split('}')[0].split(';')
        for option in options:
            option = option.strip()
            option, argument = option.split(':')
            option = option.strip()
            argument = argument.strip()
            if option == 'color':
                style['textColor'] = argument.strip()
            if option == 'background-color':
                style['backColor'] = argument.strip()

            # These two can come in any order
            if option == 'font-weight' and argument == 'bold':
                if 'fontName' in style and style['fontName'] == 'fontMonoItalic':
                    style['fontName'] = 'fontMonoBoldItalic'
                else:
                    style['fontName'] = 'fontMonoBold'
            if option == 'font-style' and argument == 'italic':
                if 'fontName' in style and style['fontName'] == 'fontSansBold':
                    style['fontName'] = 'fontMonoBoldItalic'
                else:
                    style['fontName'] = 'fontMonoItalic'
        if style.get('textColor', None) is None:
            style['textColor'] = 'black'
        styles.append([sname, style])

    # Now add default styles for all unseen class names
    for sname in classnames - seenclassnames:
        style = dstyles.get(sname, {'parent': 'code'})
        style['textColor'] = 'black'
        styles.append([sname, style])

    return dumps({'styles': styles})


for name in list(pstyles.get_all_styles()):
    css = os.popen('pygmentize -S %s -f html' % name, 'r').read()
    open(name + '.style', 'w').write(css2rl(css))



'''
This place-holder module makes the extensions directory
into a Python "package", so that external user-specific
modules can act as umbrella modules, and, for example:

import rst2pdf.extensions.preprocess_r2p

to bring in the preprocess extension.
'''



# -*- coding: utf-8 -*-
# An extension module for rst2pdf
# Copyright 2010, Patrick Maupin
# See LICENSE.txt for licensing terms

'''
preprocess is a rst2pdf extension module (invoked by ``-e preprocess``
on the rst2pdf command line).

There is a testcase for this file at tests/test_preprocess.txt

This preprocesses the source text file before handing it to docutils.

This module serves two purposes:

1) It demonstrates the technique and can be a starting point for similar
   user-written processing modules; and

2) It provides a simplified syntax for documents which are targeted only
   at rst2pdf, rather than docutils in general.

The design goal of "base rst2pdf" is to be completely compatible with
docutils, such that a file which works as a PDF can also work as HTML,
etc.

Unfortunately, base docutils is a slow-moving target, and does not
make this easy.  For example, SVG images do not work properly with
the HTML backend unless you install a patch, and docutils has no
concept of page breaks or additional vertical space (other than
the <hr>).

So, while it would be nice to have documents that render perfectly
with any backend, this goal is hard to achieve for some documents,
and once you are restricted to a particular transformation type,
then you might as well have a slightly nicer syntax for your source
document.

-----------------------------------------------------------------

Preprocessor extensions:

All current extensions except style occupy a single line in the
source file.

``.. include::``

    Processes the include file as well.  An include file may
    either be a restructured text file, OR may be an RSON or
    JSON stylesheet.  The determination is made by trying to
    parse it as RSON.  If it passes, it is a stylesheet; if not,
    well, we'll let the docutils parser have its way with it.

``.. page::``

    Is translated into a raw PageBreak.

``.. space::``

    Is translated into a raw Spacer.  If only one number given, is
    used for vertical space.  This is the canonical use case, since
    horizontal space is ignored anyway!

``.. style::``

    Allows you to create in-line stylesheets.  As with other
    restructured text components, the stylesheet data must
    be indented.  Stylesheets are in RSON or JSON.

``.. widths::``

    creates a new table style (based on table or the first
    non-numeric token) and creates a class using that style
    specifically for the next table in the document. (Creates
    a .. class::, so you must specify .. widths:: immediately
    before the table it applies to.   Allows you to set the
    widths for the table, using percentages.

``SingleWordAtLeftColumn``

    If a single word at the left column is surrounded by
    blank lines, the singleword style is automatically applied to
    the word.  This is a workaround for the broken interaction
    between docutils subtitles and bibliographic metadata.  (I
    found that docutils was referencing my subtitles from inside
    the TOC, and that seemed silly.  Perhaps there is a better
    workaround at a lower level in rst2pdf.)

-----------------------------------------------------------------

Preprocessor operation:

The preprocessor generates a file that has the same name as the source
file, with .build_temp. embedded in the name, and then passes that
file to the restructured text parser.

This file is left on the disk after operation, because any error
messages from docutils will refer to line numbers in it, rather than
in the original source, so debugging could be difficult if the
file were automatically removed.

'''

import os
import re

from rst2pdf.log import log
from rst2pdf.rson import loads as rson_loads


class DummyFile(object):
    """Stores the path and content of a file which may, or may not,
    have been written to disk.
    """

    def __init__(self, name, content):
        self.name = name
        self._content = content

    def read(self):
        return self._content


class Preprocess(object):
    def __init__(self, sourcef, incfile=False, widthcount=0):
        """Process a file and decorate the resultant Preprocess instance with
        self.result (the preprocessed file) and self.styles (extracted stylesheet
        information) for the caller.
        """

        # fix keywords dict for use by the parser.
        self.keywords = dict(
            [(x + '::', getattr(self, 'handle_' + x)) for x in self.keywords]
        )

        self.widthcount = widthcount

        name = sourcef.name
        source = sourcef.read()
        if isinstance(source, bytes):
            source = source.decode('utf8')
        source = source.replace('\r\n', '\n').replace('\r', '\n')

        # Make the determination if an include file is a stylesheet or
        # another restructured text file, and handle stylesheets appropriately.

        if incfile:
            try:
                self.styles = styles = rson_loads(source)
                substyles = styles.get('styles')
                if substyles is not None:
                    styles['styles'] = dict(substyles)
            except Exception:
                pass
            else:
                self.changed = True
                self.keep = False
                return

        # Read the whole file and wrap it in a DummyFile
        self.sourcef = DummyFile(name, source)

        # Use a regular expression on the source, to take it apart
        # and put it back together again.
        self.source = source = [x for x in self.splitter(source) if x]
        self.result = result = []
        self.styles = {}
        self.changed = False

        # More efficient to pop() a list than to keep taking tokens from [0]
        source.reverse()
        isblank = False
        keywords = self.keywords
        handle_single = keywords['single::']
        while source:
            wasblank = isblank
            isblank = False
            chunk = source.pop()
            result.append(chunk)

            # Only process single lines
            if not chunk.endswith('\n'):
                continue
            result[-1] = chunk[:-1]
            if chunk.index('\n') != len(chunk) - 1:
                continue

            # Parse the line to look for one of our keywords.
            tokens = chunk.split()
            isblank = not tokens
            if len(tokens) >= 2 and tokens[0] == '..' and tokens[1].endswith('::'):
                func = keywords.get(tokens[1])
                if func is None:
                    continue
                chunk = chunk.split('::', 1)[1]
            elif (
                wasblank
                and len(tokens) == 1
                and chunk[0].isalpha()
                and tokens[0].isalpha()
            ):
                func = handle_single
                chunk = tokens[0]
            else:
                continue

            result.pop()
            func(chunk.strip())

        # Determine if we actually did anything or not.  Just use our source file
        # if not.  Otherwise, write the results to disk (so the user can use them
        # for debugging) and return them.
        if self.changed:
            result.append('')
            result = DummyFile(name + '.build_temp', '\n'.join(result))
            self.keep = keep = len(result.read().strip())
            if keep:
                f = open(result.name, 'w')
                # Can call read a second time here because it's a DummyFile:
                f.write(result.read())
                f.close()
            self.result = result
        else:
            self.result = self.sourcef

    def handle_include(self, fname):
        # Ugly, violates DRY, etc., but I'm not about to go
        # figure out how to re-use docutils include file
        # path processing!

        for prefix in ('', os.path.dirname(self.sourcef.name)):
            try:
                f = open(os.path.join(prefix, fname), 'r')
            except IOError:
                continue
            else:
                break
        else:
            log.error("Could not find include file %s", fname)
            self.changed = True
            return

        # Recursively call this class to process include files.
        # Extract all the information from the included file.

        inc = Preprocess(f, True, self.widthcount)
        self.widthcount = inc.widthcount
        if 'styles' in self.styles and 'styles' in inc.styles:
            self.styles['styles'].update(inc.styles.pop('styles'))
        self.styles.update(inc.styles)
        if inc.changed:
            self.changed = True
            if not inc.keep:
                return
            fname = inc.result.name
        self.result.extend(['', '', '.. include:: ' + fname, ''])

    def handle_single(self, word):
        """Prepend the singleword class in front of the word."""
        self.changed = True
        self.result.extend(['', '', '.. class:: singleword', '', word, ''])

    def handle_page(self, chunk):
        """Insert a raw pagebreak"""
        self.changed = True
        self.result.extend(['', '', '.. raw:: pdf', '', '    PageBreak ' + chunk, ''])

    def handle_space(self, chunk):
        """Insert a raw space"""
        self.changed = True
        if len(chunk.replace(',', ' ').split()) == 1:
            chunk = '0 ' + chunk
        self.result.extend(['', '', '.. raw:: pdf', '', '    Spacer ' + chunk, ''])

    def handle_widths(self, chunk):
        """Insert a unique style in the stylesheet, and reference it
        from a .. class:: comment.
        """
        self.changed = True
        chunk = chunk.replace(',', ' ').replace('%', ' ').split()
        if not chunk:
            log.error('no widths specified in .. widths ::')
            return
        parent = chunk[0][0].isalpha() and chunk.pop(0) or 'table'
        values = [float(x) for x in chunk]
        total = sum(values)
        values = [int(round(100 * x / total)) for x in values]
        while 1:
            total = sum(values)
            if total > 100:
                values[values.index(max(values))] -= 1
            elif total < 100:
                values[values.index(max(values))] += 1
            else:
                break

        values = ['%s%%' % x for x in values]
        self.widthcount += 1
        stylename = 'embeddedtablewidth%d' % self.widthcount
        self.styles.setdefault('styles', {})[stylename] = dict(
            parent=parent, colWidths=values
        )
        self.result.extend(['', '', '.. class:: ' + stylename, ''])

    def handle_style(self, chunk):
        """Parse through the source until we find lines that are no longer indented,
        then pass our indented lines to the RSON parser.
        """
        self.changed = True
        if chunk:
            log.error(".. style:: does not recognize string %s" % repr(chunk))
            return

        mystyles = '\n'.join(self.read_indented())
        if not mystyles:
            log.error("Empty .. style:: block found")
        try:
            styles = rson_loads(mystyles)
        except ValueError as e:  # Error parsing the JSON data
            log.critical('Error parsing stylesheet "%s": %s' % (mystyles, str(e)))
        else:
            self.styles.setdefault('styles', {}).update(styles)

    def read_indented(self):
        """Read data from source while it is indented (or blank).
        Stop on the first non-indented line, and leave the rest
        on the source.
        """
        source = self.source
        data = None
        while source and not data:
            data = source and source.pop().splitlines() or []
            data.reverse()
            while data:
                line = data.pop().rstrip()
                if not line or line.lstrip() != line:
                    yield line
                    continue
                data.append(line)
                break
        data.reverse()
        data.append('')
        source.append('\n'.join(data))
        source.append('\n')

    # Automatically generate our keywords from methods prefixed with 'handle_'
    keywords = list(x[7:] for x in vars() if x.startswith('handle_'))

    # Generate the regular expression for parsing, and a split function using it.
    blankline = r'^([ \t]*\n)'
    singleword = r'^([A-Za-z]+[ \t]*\n)(?=[ \t]*\n)'
    comment = r'^(\.\.[ \t]+(?:%s)\:\:.*\n)' % '|'.join(keywords)
    expression = '(?:%s)' % '|'.join([blankline, singleword, comment])
    splitter = re.compile(expression, re.MULTILINE).split


class MyStyles(str):
    """This class conforms to the styles.py processing requirements
    for a stylesheet that is not really a file.  It must be callable(),
    and str(x) must return the name of the stylesheet.
    """

    def __new__(cls, styles):
        self = str.__new__(cls, 'Embedded Preprocess Styles')
        self.data = styles
        return self

    def __call__(self):
        return self.data


def install(createpdf, options):
    """This is where we intercept the document conversion.
    Preprocess the restructured text, and insert our
    new styles (if any).
    """
    data = Preprocess(options.infile)
    options.infile = data.result
    if data.styles:
        options.style.append(MyStyles(data.styles))



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

# Some fragments of code are copied from Reportlab under this license:
#
#####################################################################################
#
#       Copyright (c) 2000-2008, ReportLab Inc.
#       All rights reserved.
#
#       Redistribution and use in source and binary forms, with or without modification,
#       are permitted provided that the following conditions are met:
#
#               *       Redistributions of source code must retain the above copyright notice,
#                       this list of conditions and the following disclaimer.
#               *       Redistributions in binary form must reproduce the above copyright notice,
#                       this list of conditions and the following disclaimer in the documentation
#                       and/or other materials provided with the distribution.
#               *       Neither the name of the company nor the names of its contributors may be
#                       used to endorse or promote products derived from this software without
#                       specific prior written permission.
#
#       THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#       ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#       WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#       IN NO EVENT SHALL THE OFFICERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#       INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
#       TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#       OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
#       IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
#       IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#       SUCH DAMAGE.
#
#####################################################################################

from copy import copy

import docutils.nodes
import reportlab
from reportlab.platypus import Paragraph, TableStyle
from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT

from .basenodehandler import NodeHandler
from .directives.aafigure import Aanode
from .directives.oddeven import OddEvenNode
from .flowables import (
    Table,
    DelayedTable,
    SplitTable,
    Heading,
    MyTableOfContents,
    MySpacer,
    Separation,
    BoxedContainer,
    MyPageBreak,
    tablepadding,
    OddEven,
    XPreformatted,
)
from .math_flowable import Math
from .utils import parseRaw, parseHTML


class TocBuilderVisitor(docutils.nodes.SparseNodeVisitor):
    def __init__(self, document):
        docutils.nodes.SparseNodeVisitor.__init__(self, document)
        self.toc = None
        # For some reason, when called via sphinx,
        # .. contents:: ends up trying to call
        # visitor.document.reporter.debug
        # so we need a valid document here.
        self.document = docutils.utils.new_document('')

    def visit_reference(self, node):
        refid = node.attributes.get('refid')
        if refid:
            self.toc.refids.append(refid)


class HandleDocument(NodeHandler, docutils.nodes.document):
    pass


class HandleTable(NodeHandler, docutils.nodes.table):
    def gather_elements(self, client, node, style):
        if node['classes']:
            style = client.styles.combinedStyle(['table'] + node['classes'])
        else:
            style = client.styles['table']
        return (
            [MySpacer(0, client.styles['table'].spaceBefore)]
            + client.gather_elements(node, style=style)
            + [MySpacer(0, client.styles['table'].spaceAfter)]
        )


class HandleTGroup(NodeHandler, docutils.nodes.tgroup):
    def gather_elements(self, client, node, style):

        # Take the style from the parent "table" node
        # because sometimes it's not passed down.

        if node.parent['classes']:
            style = client.styles.combinedStyle(['table'] + node.parent['classes'])
        else:
            style = client.styles['table']
        rows = []
        colWidths = []
        hasHead = False
        headRows = 0
        for n in node.children:
            if isinstance(n, docutils.nodes.thead):
                hasHead = True
                for row in n.children:
                    r = []
                    for cell in row.children:
                        r.append(cell)
                    rows.append(r)
                headRows = len(rows)
            elif isinstance(n, docutils.nodes.tbody):
                for row in n.children:
                    r = []
                    for cell in row.children:
                        r.append(cell)
                    rows.append(r)
            elif isinstance(n, docutils.nodes.colspec):
                colWidths.append(int(n['colwidth']))

        # colWidths are in no specific unit, really. Maybe ems.
        tot = sum(colWidths)
        adjustedWidths = []
        for x in colWidths:
            # Convert them to %
            w = 100 * x / tot
            # Limit minimum width of a column as narrow columns cause strange "huge height" from reportlab if the cell
            # padding is larger than the calculated width. Hack for #967
            if w < 4:
                w = 4
            adjustedWidths.append("%s%%" % w)

        colWidths = adjustedWidths

        if 'colWidths' in style.__dict__:
            colWidths[: len(style.colWidths)] = style.colWidths

        spans = client.filltable(rows)

        data = []
        rowids = range(0, len(rows))
        for row, i in zip(rows, rowids):
            r = []
            j = 0
            for cell in row:
                if isinstance(cell, str):
                    r.append("")
                else:
                    if i < headRows:
                        st = client.styles['table-heading']
                    else:
                        st = client.styles['table-body']
                    ell = client.gather_elements(cell, style=st)
                    r.append(ell)
                j += 1
            data.append(r)

        st = TableStyle(spans)
        if 'commands' in style.__dict__:
            for cmd in style.commands:
                st.add(*cmd)
        else:
            # Only use the commands from "table" if the
            # specified class has no commands.

            for cmd in client.styles['table'].commands:
                st.add(*cmd)

        if hasHead:
            for cmd in client.styles.tstyleHead(headRows):
                st.add(*cmd)
        rtr = client.repeat_table_rows

        t = DelayedTable(data, colWidths, st, rtr)
        if style.alignment == TA_LEFT:
            t.hAlign = 'LEFT'
        elif style.alignment == TA_CENTER:
            t.hAlign = 'CENTER'
        elif style.alignment == TA_RIGHT:
            t.hAlign = 'RIGHT'
        return [t]


class HandleParagraph(NodeHandler, docutils.nodes.paragraph):
    def gather_elements(self, client, node, style):
        return [Paragraph(client.gen_pdftext(node), style)]

    def get_pre_post(self, client, node, replaceEnt):
        pre = ''
        targets = set(node.get('ids', []) + client.pending_targets)
        client.pending_targets = []
        for _id in targets:
            if _id not in client.targets:
                pre += '<a name="%s"/>' % (_id)
                client.targets.append(_id)
        return pre, '\n'


class HandleTitle(HandleParagraph, docutils.nodes.title):
    def gather_elements(self, client, node, style):
        # Special cases: (Not sure this is right ;-)
        if isinstance(node.parent, docutils.nodes.document):
            # The visible output is done by the cover template
            node.elements = []
            client.doc_title = node.rawsource
            client.doc_title_clean = node.astext().strip()
        elif isinstance(node.parent, docutils.nodes.topic):
            node.elements = [
                Paragraph(client.gen_pdftext(node), client.styles['topic-title'])
            ]
        elif isinstance(node.parent, docutils.nodes.Admonition):
            node.elements = [
                Paragraph(client.gen_pdftext(node), client.styles['admonition-title'])
            ]
        elif isinstance(node.parent, docutils.nodes.table):
            node.elements = [
                Paragraph(client.gen_pdftext(node), client.styles['table-title'])
            ]
        elif isinstance(node.parent, docutils.nodes.sidebar):
            node.elements = [
                Paragraph(client.gen_pdftext(node), client.styles['sidebar-title'])
            ]
        else:
            # Section/Subsection/etc.
            text = client.gen_pdftext(node)
            maxdepth = 6
            # The parent ID is the refid + an ID to make it unique for Sphinx
            parent_id = (
                (node.parent.get('ids', [None]) or [None])[0] + u'-' + str(id(node))
            )
            node.elements = [
                Heading(
                    text,
                    client.styles['heading%d' % min(client.depth, maxdepth)],
                    level=client.depth - 1,
                    parent_id=parent_id,
                    node=node,
                    section_header_depth=client.section_header_depth,
                    toc_depth=client.toc_depth,
                )
            ]
            if client.depth <= client.breaklevel:
                node.elements.insert(0, MyPageBreak(breakTo=client.breakside))
        return node.elements


class HandleSubTitle(HandleParagraph, docutils.nodes.subtitle):
    def gather_elements(self, client, node, style):
        if isinstance(node.parent, docutils.nodes.sidebar):
            elements = [
                Paragraph(client.gen_pdftext(node), client.styles['sidebar-subtitle'])
            ]
        elif isinstance(node.parent, docutils.nodes.document):
            # elements = [Paragraph(client.gen_pdftext(node),
            # client.styles['subtitle'])]
            # The visible output is now done by the cover template
            elements = []
            # FIXME: looks like subtitles don't have a rawsource like
            # titles do.
            # That means that literals and italics etc in subtitles won't
            # work.
            client.doc_subtitle = getattr(node, 'rawtext', node.astext()).strip()
        else:
            elements = [
                Paragraph(
                    client.gen_pdftext(node),
                    client.styles['%s-subtitle' % node.parent.tagname],
                )
            ]

        return elements


class HandleDocInfo(NodeHandler, docutils.nodes.docinfo):
    # A docinfo usually contains several fields.
    # We'll render it as a series of elements, one field each.
    pass


class HandleField(NodeHandler, docutils.nodes.field):
    def gather_elements(self, client, node, style):
        # A field has two child elements, a field_name and a field_body.
        # We render as a two-column table, left-column is right-aligned,
        # bold, and much smaller

        fn = Paragraph(
            client.gather_pdftext(node.children[0]) + ":",
            style=client.styles['fieldname'],
        )
        fb = client.gen_elements(node.children[1], style=client.styles['fieldvalue'])
        t_style = TableStyle(client.styles['field-list'].commands)
        return [
            DelayedTable(
                [[fn, fb]],
                colWidths=client.styles['field-list'].colWidths,
                style=t_style,
            )
        ]


class HandleDecoration(NodeHandler, docutils.nodes.decoration):
    pass


class HandleHeader(NodeHandler, docutils.nodes.header):
    stylename = 'header'

    def gather_elements(self, client, node, style):
        client.decoration[self.stylename] = client.gather_elements(
            node, style=client.styles[self.stylename]
        )
        return []


class HandleFooter(HandleHeader, docutils.nodes.footer):
    stylename = 'footer'


class HandleAuthor(NodeHandler, docutils.nodes.author):
    def gather_elements(self, client, node, style):
        if isinstance(node.parent, docutils.nodes.authors):
            # Is only one of multiple authors. Return a paragraph
            node.elements = [Paragraph(client.gather_pdftext(node), style=style)]
            if client.doc_author:
                client.doc_author += (
                    client.author_separator(style=style) + node.astext().strip()
                )
            else:
                client.doc_author = node.astext().strip()
        else:
            # A single author: works like a field
            fb = client.gather_pdftext(node)

            t_style = TableStyle(client.styles['field-list'].commands)
            colWidths = [
                client.styles.adjustUnits(x)
                for x in client.styles['field-list'].colWidths
            ]

            node.elements = [
                Table(
                    [
                        [
                            Paragraph(
                                client.text_for_label("author", style) + ":",
                                style=client.styles['fieldname'],
                            ),
                            Paragraph(fb, style),
                        ]
                    ],
                    style=t_style,
                    colWidths=colWidths,
                )
            ]
            client.doc_author = node.astext().strip()
        return node.elements


class HandleAuthors(NodeHandler, docutils.nodes.authors):
    def gather_elements(self, client, node, style):
        # Multiple authors. Create a two-column table.
        # Author references on the right.
        t_style = TableStyle(client.styles['field-list'].commands)
        colWidths = client.styles['field-list'].colWidths

        td = [
            [
                Paragraph(
                    client.text_for_label("authors", style) + ":",
                    style=client.styles['fieldname'],
                ),
                client.gather_elements(node, style=style),
            ]
        ]
        return [DelayedTable(td, colWidths=colWidths, style=t_style)]


class HandleFList(NodeHandler):
    adjustwidths = False
    TableType = DelayedTable

    def gather_elements(self, client, node, style):
        fb = client.gather_pdftext(node)
        t_style = TableStyle(client.styles['field-list'].commands)
        colWidths = client.styles['field-list'].colWidths
        if self.adjustwidths:
            colWidths = [client.styles.adjustUnits(x) for x in colWidths]
        label = client.text_for_label(self.labeltext, style) + ":"
        t = self.TableType(
            [
                [
                    Paragraph(label, style=client.styles['fieldname']),
                    Paragraph(fb, style),
                ]
            ],
            style=t_style,
            colWidths=colWidths,
        )
        return [t]


class HandleOrganization(HandleFList, docutils.nodes.organization):
    labeltext = "organization"


class HandleContact(HandleFList, docutils.nodes.contact):
    labeltext = "contact"


class HandleAddress(HandleFList, docutils.nodes.address):
    labeltext = "address"

    def gather_elements(self, client, node, style):
        fb = client.gather_pdftext(node)
        t_style = TableStyle(client.styles['field-list'].commands)
        colWidths = client.styles['field-list'].colWidths
        if self.adjustwidths:
            colWidths = [client.styles.adjustUnits(x) for x in colWidths]
        label = client.text_for_label(self.labeltext, style) + ":"
        t = self.TableType(
            [
                [
                    Paragraph(label, style=client.styles['fieldname']),
                    XPreformatted(fb, style),
                ]
            ],
            style=t_style,
            colWidths=colWidths,
        )
        return [t]


class HandleVersion(HandleFList, docutils.nodes.version):
    labeltext = "version"


class HandleRevision(HandleFList, docutils.nodes.revision):
    labeltext = "revision"
    adjustwidths = True
    TableType = Table


class HandleStatus(HandleFList, docutils.nodes.status):
    labeltext = "status"


class HandleDate(HandleFList, docutils.nodes.date):
    labeltext = "date"


class HandleCopyright(HandleFList, docutils.nodes.copyright):
    labeltext = "copyright"


class HandleTopic(NodeHandler, docutils.nodes.topic):
    def gather_elements(self, client, node, style):
        # toc
        node_classes = node.attributes.get('classes', [])
        cstyles = client.styles
        if 'contents' in node_classes:
            toc_visitor = TocBuilderVisitor(node.document)
            if 'local' in node_classes:
                toc_visitor.toc = MyTableOfContents(parent=node.parent)
            else:
                toc_visitor.toc = MyTableOfContents(parent=None)
            toc_visitor.toc.linkColor = cstyles.tocColor or cstyles.linkColor
            node.walk(toc_visitor)
            toc = toc_visitor.toc
            toc.levelStyles = [cstyles['toc%d' % l] for l in range(1, 15)]
            for s in toc.levelStyles:
                # FIXME: awful slimy hack!
                s.__class__ = reportlab.lib.styles.ParagraphStyle

            if 'local' in node_classes:
                node.elements = [toc]
            else:
                node.elements = [
                    Paragraph(
                        client.gen_pdftext(node.children[0]),
                        cstyles['heading1'],
                    ),
                    toc,
                ]
        else:
            node.elements = client.gather_elements(node, style=style)
        return node.elements


class HandleFieldBody(NodeHandler, docutils.nodes.field_body):
    pass


class HandleSection(NodeHandler, docutils.nodes.section):
    def gather_elements(self, client, node, style):
        # XXX: should style be passed down here?
        client.depth += 1
        elements = client.gather_elements(node)
        client.depth -= 1
        return elements


class HandleBulletList(NodeHandler, docutils.nodes.bullet_list):
    def gather_elements(self, client, node, style):

        if node['classes']:
            style = client.styles[node['classes'][0]]
        else:
            style = client.styles["bullet-list"]

        node.elements = client.gather_elements(node, style=style)

        # Here we need to separate the list from the previous element.
        # Calculate by how much:

        sb = style.spaceBefore  # list separation
        sa = style.spaceAfter  # list separation

        node.elements.insert(0, MySpacer(0, sb))
        node.elements.append(MySpacer(0, sa))
        return node.elements


class HandleDefOrOptList(
    NodeHandler, docutils.nodes.definition_list, docutils.nodes.option_list
):
    pass


class HandleFieldList(NodeHandler, docutils.nodes.field_list):
    def gather_elements(self, client, node, style):
        return [
            MySpacer(0, client.styles['field-list'].spaceBefore)
        ] + client.gather_elements(node, style=style)


class HandleEnumeratedList(NodeHandler, docutils.nodes.enumerated_list):
    def gather_elements(self, client, node, style):
        if node['classes']:
            style = client.styles[node['classes'][0]]
        else:
            style = client.styles["item-list"]

        node.elements = client.gather_elements(node, style=style)

        # Here we need to separate the list from the previous element.
        # Calculate by how much:

        sb = style.spaceBefore  # list separation
        sa = style.spaceAfter  # list separation

        node.elements.insert(0, MySpacer(0, sb))
        node.elements.append(MySpacer(0, sa))
        return node.elements


class HandleDefinition(NodeHandler, docutils.nodes.definition):
    def gather_elements(self, client, node, style):
        return client.gather_elements(node, style=style)


class HandleOptionListItem(NodeHandler, docutils.nodes.option_list_item):
    def gather_elements(self, client, node, style):
        optext = ', '.join(
            [client.gather_pdftext(child) for child in node.children[0].children]
        )

        desc = client.gather_elements(node.children[1], style)

        t_style = TableStyle(client.styles['option-list'].commands)
        colWidths = client.styles['option-list'].colWidths
        node.elements = [
            DelayedTable(
                [[client.PreformattedFit(optext, client.styles["literal"]), desc]],
                colWidths=colWidths,
                style=t_style,
            )
        ]
        return node.elements


class HandleDefListItem(NodeHandler, docutils.nodes.definition_list_item):
    def gather_elements(self, client, node, style):
        # I need to catch the classifiers here
        tt = []
        dt = []
        ids = []
        for n in node.children:
            if isinstance(n, docutils.nodes.term):
                for i in n['ids']:  # Used by sphinx glossary lists
                    if i not in client.targets:
                        ids.append('<a name="%s"/>' % i)
                        client.targets.append(i)
                o, c = client.styleToTags("definition-list-term")
                tt.append(o + client.gather_pdftext(n) + c)
            elif isinstance(n, docutils.nodes.classifier):
                o, c = client.styleToTags("definition-list-classifier")
                tt.append(o + client.gather_pdftext(n) + c)
            else:
                dt.extend(client.gen_elements(n, style))

        # FIXME: make this configurable from the stylesheet
        t_style = TableStyle(client.styles['definition'].commands)
        cw = getattr(client.styles['definition'], 'colWidths', [])

        if client.splittables:
            node.elements = [
                Paragraph(
                    ''.join(ids) + ' : '.join(tt),
                    client.styles['definition-list-term'],
                ),
                SplitTable([['', dt]], colWidths=cw, style=t_style),
            ]
        else:
            node.elements = [
                Paragraph(
                    ''.join(ids) + ' : '.join(tt),
                    client.styles['definition-list-term'],
                ),
                DelayedTable([['', dt]], colWidths=[10, None], style=t_style),
            ]

        return node.elements


class HandleListItem(NodeHandler, docutils.nodes.list_item):
    def gather_elements(self, client, node, style):
        b, t = client.bullet_for_node(node)

        bStyle = copy(style)
        bStyle.alignment = 2

        # FIXME: use different unicode bullets depending on b
        if b and b in "*+-":
            b = getattr(bStyle, 'bulletText', u'\u2022')

        # The style has information about the bullet:
        #
        # bulletFontSize
        # bulletFont
        # This is so the baselines of the bullet and the text align
        extra_space = bStyle.bulletFontSize - bStyle.fontSize

        bStyle.fontSize = bStyle.bulletFontSize
        bStyle.fontName = bStyle.bulletFontName

        if t == 'bullet':
            item_st = client.styles['bullet-list-item']
        else:
            item_st = client.styles['item-list-item']

        el = client.gather_elements(node, item_st)
        # FIXME: this is really really not good code
        if not el:
            el = [Paragraph(u"<nobr>\xa0</nobr>", item_st)]

        idx = node.parent.children.index(node)
        if idx == 0:
            # The first item in the list, so doesn't need
            # separation (it's provided by the list itself)
            sb = 0
        else:
            # Not the first item, so need to separate from
            # previous item. Account for space provided by
            # the item's content, too.
            sb = item_st.spaceBefore - item_st.spaceAfter

        bStyle.spaceBefore = 0

        t_style = TableStyle(style.commands)
        # The -3 here is to compensate for padding, 0 doesn't work :-(
        t_style._cmds.extend([["BOTTOMPADDING", [0, 0], [-1, -1], -3]])
        if extra_space > 0:
            # The bullet is larger, move down the item text
            sb += extra_space

        colWidths = getattr(style, 'colWidths', [])
        while len(colWidths) < 2:
            colWidths.append(client.styles['item_list'].colWidths[len(colWidths)])

        if client.splittables:
            node.elements = [
                MySpacer(0, sb),
                SplitTable(
                    [[Paragraph(b, style=bStyle), el]],
                    colWidths=colWidths,
                    style=t_style,
                ),
            ]
        else:
            node.elements = [
                MySpacer(0, sb),
                DelayedTable(
                    [[Paragraph(b, style=bStyle), el]],
                    colWidths=colWidths,
                    style=t_style,
                ),
            ]
        return node.elements


class HandleTransition(NodeHandler, docutils.nodes.transition):
    def gather_elements(self, client, node, style):
        return [Separation(style=client.styles['separation'])]


class HandleBlockQuote(NodeHandler, docutils.nodes.block_quote):
    def gather_elements(self, client, node, style):
        # This should work, but doesn't look good inside of
        # table cells (see Issue 173)
        # node.elements = [MyIndenter(left=client.styles['blockquote'].leftIndent)]\
        # + client.gather_elements( node, style) + \
        # [MyIndenter(left=-client.styles['blockquote'].leftIndent)]
        # Workaround for Issue 173 using tables
        leftIndent = client.styles['blockquote'].leftIndent
        rightIndent = client.styles['blockquote'].rightIndent
        spaceBefore = client.styles['blockquote'].spaceBefore
        spaceAfter = client.styles['blockquote'].spaceAfter
        s = copy(client.styles['blockquote'])
        s.leftIndent = style.leftIndent
        data = [['', client.gather_elements(node, s)]]
        if client.splittables:
            node.elements = [
                MySpacer(0, spaceBefore),
                SplitTable(
                    data,
                    colWidths=[leftIndent, None],
                    style=TableStyle(
                        [
                            ["TOPPADDING", [0, 0], [-1, -1], 0],
                            ["LEFTPADDING", [0, 0], [-1, -1], 0],
                            ["RIGHTPADDING", [0, 0], [-1, -1], rightIndent],
                            ["BOTTOMPADDING", [0, 0], [-1, -1], 0],
                        ]
                    ),
                ),
                MySpacer(0, spaceAfter),
            ]
        else:
            node.elements = [
                MySpacer(0, spaceBefore),
                DelayedTable(
                    data,
                    colWidths=[leftIndent, None],
                    style=TableStyle(
                        [
                            ["TOPPADDING", [0, 0], [-1, -1], 0],
                            ["LEFTPADDING", [0, 0], [-1, -1], 0],
                            ["RIGHTPADDING", [0, 0], [-1, -1], rightIndent],
                            ["BOTTOMPADDING", [0, 0], [-1, -1], 0],
                        ]
                    ),
                ),
                MySpacer(0, spaceAfter),
            ]
        return node.elements


class HandleAttribution(NodeHandler, docutils.nodes.attribution):
    def gather_elements(self, client, node, style):
        return [Paragraph(client.gather_pdftext(node), client.styles['attribution'])]


class HandleComment(NodeHandler, docutils.nodes.comment):
    def gather_elements(self, client, node, style):
        # Class that generates no output
        return []


class HandleLineBlock(NodeHandler, docutils.nodes.line_block):
    def gather_elements(self, client, node, style):
        if isinstance(node.parent, docutils.nodes.line_block):
            qstyle = copy(style)
            qstyle.leftIndent += client.styles.adjustUnits("1.5em")
        else:
            qstyle = copy(client.styles['lineblock'])
        # Fix Issue 225: no space betwen line in a lineblock, but keep
        # space before the lineblock itself
        # Fix Issue 482: nested lineblocks don't need spacing before/after
        if not isinstance(node.parent, docutils.nodes.line_block):
            return (
                [MySpacer(0, client.styles['lineblock'].spaceBefore)]
                + client.gather_elements(node, style=qstyle)
                + [MySpacer(0, client.styles['lineblock'].spaceAfter)]
            )
        else:
            return client.gather_elements(node, style=qstyle)


class HandleLine(NodeHandler, docutils.nodes.line):
    def gather_elements(self, client, node, style):
        # line nodes have no classes, they have to inherit from the outermost lineblock (sigh)
        # For more info see Issue 471 and its test case.

        parent = node
        while isinstance(
            parent.parent, (docutils.nodes.line, docutils.nodes.line_block)
        ):
            parent = parent.parent
        p_class = (parent.get('classes') or ['line'])[0]
        qstyle = copy(client.styles[p_class])
        # Indent .5em per indent unit
        i = node.__dict__.get('indent', 0)
        # qstyle = copy(client.styles['line'])
        qstyle.leftIndent += client.styles.adjustUnits("0.5em") * i
        text = client.gather_pdftext(node)
        if not text:  # empty line
            text = u"<nobr>\xa0</nobr>"
        return [Paragraph(text, style=qstyle)]


class HandleLiteralBlock(
    NodeHandler, docutils.nodes.literal_block, docutils.nodes.doctest_block
):
    def gather_elements(self, client, node, style):
        if node['classes']:
            style = client.styles.combinedStyle(['code'] + node['classes'])
        else:
            style = client.styles['code']

        return [
            client.PreformattedFit(client.gather_pdftext(node, replaceEnt=True), style)
        ]


class HandleFigure(NodeHandler, docutils.nodes.figure):
    def gather_elements(self, client, node, style):

        # Either use the figure style or the class
        # selected by the user
        st_name = 'figure'
        if node.get('classes'):
            st_name = node.get('classes')[0]
        style = client.styles[st_name]
        cmd = getattr(style, 'commands', [])

        w = node.get('width', client.styles['figure'].colWidths[0])
        cw = [w]
        sub_elems = client.gather_elements(node, style=None)
        t_style = TableStyle(cmd)
        table = DelayedTable([[e] for e in sub_elems], colWidths=cw, style=t_style)
        table.hAlign = node.get('align', 'CENTER').upper()
        return [
            MySpacer(0, style.spaceBefore),
            table,
            MySpacer(0, style.spaceAfter),
        ]


class HandleCaption(NodeHandler, docutils.nodes.caption):
    def gather_elements(self, client, node, style):
        return [
            Paragraph(
                client.gather_pdftext(node),
                style=client.styles['figure-caption'],
            )
        ]


class HandleLegend(NodeHandler, docutils.nodes.legend):
    def gather_elements(self, client, node, style):
        return client.gather_elements(node, style=client.styles['figure-legend'])


class HandleSidebar(NodeHandler, docutils.nodes.sidebar):
    def gather_elements(self, client, node, style):
        return [
            BoxedContainer(
                client.gather_elements(node, style=None),
                client.styles['sidebar'],
            )
        ]


class HandleRubric(NodeHandler, docutils.nodes.rubric):
    def gather_elements(self, client, node, style):
        # Sphinx uses a rubric as footnote container
        if (
            self.sphinxmode
            and len(node.children) == 1
            and node.children[0].astext() == 'Footnotes'
        ):
            return []
        else:
            return [Paragraph(client.gather_pdftext(node), client.styles['rubric'])]


class HandleCompound(NodeHandler, docutils.nodes.compound):
    # FIXME think if this is even implementable
    pass


class HandleContainer(NodeHandler, docutils.nodes.container):
    def getelements(self, client, node, style):
        parent = node.parent
        if not isinstance(parent, (docutils.nodes.header, docutils.nodes.footer)):
            return NodeHandler.getelements(self, client, node, style)
        return self.gather_elements(client, node, style)


class HandleSubstitutionDefinition(NodeHandler, docutils.nodes.substitution_definition):
    def gather_elements(self, client, node, style):
        return []


class HandleTBody(NodeHandler, docutils.nodes.tbody):
    def gather_elements(self, client, node, style):
        rows = [client.gen_elements(n) for n in node.children]
        t = []
        for r in rows:
            if not r:
                continue
            t.append(r)
        t_style = TableStyle(client.styles['table'].commands)
        colWidths = client.styles['table'].colWidths
        return [DelayedTable(t, colWidths=colWidths, style=t_style)]


class HandleFootnote(NodeHandler, docutils.nodes.footnote, docutils.nodes.citation):
    def gather_elements(self, client, node, style):
        # It seems a footnote contains a label and a series of elements
        ltext = client.gather_pdftext(node.children[0])
        label = None
        ids = ''
        for i in node.get('ids', []):
            ids += '<a name="%s"/>' % (i)
        client.targets.extend(node.get('ids', [ltext]))

        if len(node['backrefs']) > 1 and client.footnote_backlinks:
            backrefs = []
            i = 1
            for r in node['backrefs']:
                backrefs.append(
                    '<a href="#%s" color="%s">%d</a>' % (r, client.styles.linkColor, i)
                )
                i += 1
            backrefs = '(%s)' % ', '.join(backrefs)
            if ltext not in client.targets:
                label = Paragraph(
                    ids + '%s' % (ltext + backrefs), client.styles["endnote"]
                )
                client.targets.append(ltext)
        elif len(node['backrefs']) == 1 and client.footnote_backlinks:
            if ltext not in client.targets:
                label = Paragraph(
                    ids
                    + '<a href="#%s" color="%s">%s</a>'
                    % (node['backrefs'][0], client.styles.linkColor, ltext),
                    client.styles["endnote"],
                )
                client.targets.append(ltext)
        else:
            if ltext not in client.targets:
                label = Paragraph(ids + ltext, client.styles["endnote"])
                client.targets.append(ltext)
        if not label:
            label = Paragraph(ids + ltext, client.styles["endnote"])
        contents = client.gather_elements(node, client.styles["endnote"])[1:]
        if client.inline_footnotes:
            st = client.styles['endnote']
            t_style = TableStyle(st.commands)
            colWidths = client.styles['endnote'].colWidths
            node.elements = [
                MySpacer(0, st.spaceBefore),
                DelayedTable([[label, contents]], colWidths=colWidths, style=t_style),
                MySpacer(0, st.spaceAfter),
            ]
            if client.real_footnotes:
                client.mustMultiBuild = True
                for e in node.elements:
                    e.isFootnote = True
        else:
            client.decoration['endnotes'].append([label, contents])
            node.elements = []
        return node.elements


class HandleLabel(NodeHandler, docutils.nodes.label):
    def gather_elements(self, client, node, style):
        return [Paragraph(client.gather_pdftext(node), style)]


class HandleEntry(NodeHandler, docutils.nodes.entry):
    pass


class HandleRaw(NodeHandler, docutils.nodes.raw):
    def gather_elements(self, client, node, style):
        # Not really raw, but what the heck
        if node.get('format', 'NONE').lower() == 'pdf':
            return parseRaw(str(node.astext()), node)
        elif client.raw_html and node.get('format', 'NONE').lower() == 'html':
            x = parseHTML(str(node.astext()), node)
            return x
        else:
            return []


class HandleOddEven(NodeHandler, OddEvenNode):
    def gather_elements(self, client, node, style):
        odd = []
        even = []
        if node.children:
            if isinstance(node.children[0], docutils.nodes.paragraph):
                if node.children[0].get('classes'):
                    s = client.styles[node.children[0].get('classes')[0]]
                else:
                    s = style
                odd = [Paragraph(client.gather_pdftext(node.children[0]), s)]
            else:
                # A compound element
                odd = client.gather_elements(node.children[0])
        if len(node.children) > 1:
            if isinstance(node.children[1], docutils.nodes.paragraph):
                if node.children[1].get('classes'):
                    s = client.styles[node.children[1].get('classes')[0]]
                else:
                    s = style
                even = [Paragraph(client.gather_pdftext(node.children[1]), s)]
            else:
                even = client.gather_elements(node.children[1])

        return [OddEven(odd=odd, even=even)]


class HandleAanode(NodeHandler, Aanode):
    def gather_elements(self, client, node, style):
        style_options = {
            'font': client.styles['aafigure'].fontName,
        }
        return [node.gen_flowable(style_options)]


class HandleAdmonition(
    NodeHandler,
    docutils.nodes.attention,
    docutils.nodes.caution,
    docutils.nodes.danger,
    docutils.nodes.error,
    docutils.nodes.hint,
    docutils.nodes.important,
    docutils.nodes.note,
    docutils.nodes.tip,
    docutils.nodes.warning,
    docutils.nodes.Admonition,
):
    def gather_elements(self, client, node, style):
        if node.children and isinstance(node.children[0], docutils.nodes.title):
            title = []
        else:
            title = [
                Paragraph(
                    client.text_for_label(node.tagname, style),
                    style=client.styles['%s-heading' % node.tagname],
                )
            ]
        rows = title + client.gather_elements(node, style=style)
        st = client.styles[node.tagname]
        if 'commands' in dir(st):
            t_style = TableStyle(st.commands)
        else:
            t_style = TableStyle()
        t_style.add("ROWBACKGROUNDS", [0, 0], [-1, -1], [st.backColor])
        t_style.add("BOX", [0, 0], [-1, -1], st.borderWidth, st.borderColor)

        if client.splittables:
            node.elements = [
                MySpacer(0, st.spaceBefore),
                SplitTable(
                    [['', rows]],
                    colWidths=[0, None],
                    style=t_style,
                    padding=st.borderPadding,
                ),
                MySpacer(0, st.spaceAfter),
            ]
        else:
            padding, p1, p2, p3, p4 = tablepadding(padding=st.borderPadding)
            t_style.add(*p1)
            t_style.add(*p2)
            t_style.add(*p3)
            t_style.add(*p4)
            node.elements = [
                MySpacer(0, st.spaceBefore),
                DelayedTable([['', rows]], colWidths=[0, None], style=t_style),
                MySpacer(0, st.spaceAfter),
            ]
        return node.elements


class HandleMath(NodeHandler, docutils.nodes.math_block, docutils.nodes.math):
    def gather_elements(self, client, node, style):
        label = node.attributes.get('label')
        return [
            Math(
                node.astext(),
                label,
                style,
            )
        ]

    def get_text(self, client, node, replaceEnt):
        """Generate an image of the math equation for the inline role usage"""
        # get style for current node
        style = client.styles.styleForNode(node)
        label = node.attributes.get('label')
        mf = Math(node.astext(), label, style)
        mf.fontsize = style.fontSize * 0.95
        w, h = mf.wrap(0, 0)
        descent = mf.descent()
        img = mf.genImage()
        client.to_unlink.append(img)
        return '<img src="%s" width="%f" height="%f" valign="%f"/>' % (
            img,
            w,
            h,
            -descent,
        )



"""
RSON -- readable serial object notation

RSON is a superset of JSON with relaxed syntax for human readability.

Simple usage example:
            import rson
            obj = rson.loads(source)

Additional documentation available at:

http://code.google.com/p/rson/
"""

__version__ = '0.08'

__author__ = 'Patrick Maupin <pmaupin@gmail.com>'

__copyright__ = '''
Copyright (c) 2010, Patrick Maupin.  All rights reserved.

 Permission is hereby granted, free of charge, to any person
 obtaining a copy of this software and associated documentation
 files (the "Software"), to deal in the Software without
 restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following
 conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.
 '''

import bisect
import re
import sys


class RSONDecodeError(ValueError):
    pass


class Tokenizer(list):
    """The RSON tokenizer uses re.split() to rip the source string
    apart into smaller strings which may or may not be true standalone
    tokens.  Sufficient information is maintained to put Humpty-Dumpty
    back together when necessary.

    The tokens are kept in a reversed list.  This makes retrieving
    the next token a low-cost pop() operation from the end of the list,
    and facilitates arbitrary lookahead operations.

    Each token is a tuple, containing the following elements:
       [0] Negative character offset of token within the source string.
           A negative offset is used so that the tokens are sorted
           properly for bisect() for special operations.
       [1] single-character string usually containing a character
           which represents the token type (often the entire token).
       [2] string containing entire token
       [3] string (possibly null) containing whitespace after token
       [4] Indentation value of line containing token
                        (\n followed by 0 or more spaces)
       [5] Line number of token
       [6] Tokenizer object that the token belongs to (for error reporting)
    """

    # Like Python, indentation is special.  I originally planned on making
    # the space character the only valid indenter, but that got messy
    # when combined with the desire to be 100% JSON compatible, so, like
    # JSON, you can indent with any old whitespace, but if you mix and
    # match, you might be in trouble (like with Python).
    #
    # An indentation is always the preceding EOL plus optional spaces,
    # so we create a dummy EOL for the very start of the string.
    # Could also have an embedded comment
    indentation = r'\n[ \t\v\f]*(?:#.*)?'

    # RSON syntax delimiters are tokenized separately from everything else.
    delimiterset = set(' { } [ ] : = , '.split())

    re_delimiterset = ''.join(delimiterset).replace(']', r'\]').replace('[', r'\[')

    # Create a RE pattern for the delimiters
    delimiter_pattern = '[%s]' % re_delimiterset

    # A regular quoted string must terminate before the end of the line,
    # and \ can be used as the internal escape character.
    quoted_string = r'"(?:[^"\n\\]|\\.)*(?:"|(?=\n))'

    # A triple-quoted string can contain any characters.  The only escape
    # processing that is done on them is to allow a backslash in front of
    # another set of triple quotes.  We only look for the start of one of
    # these suckers in the first pass, and then go find the real string
    # later.  This keeps us from getting our knickers in a twist on
    # regular strings vs triple-quoted vs comments, etc.

    triple_quoted_string = '"""'

    # Any non-whitespace, non-delimiter, group of characters is in the "other"
    # category.  This group can have embedded whitespace, but ends on a
    # non-whitespace character.

    other = r'[\S](?:[^%s\n]*[^%s\s])*' % (re_delimiterset, re_delimiterset)

    pattern = '(%s)' % '|'.join(
        [delimiter_pattern, triple_quoted_string, quoted_string, other, indentation]
    )

    splitter = re.compile(pattern).split

    @classmethod
    def factory(cls):
        splitter = cls.splitter
        delimiterset = set(cls.delimiterset) | set('"')

        def newstring(source, client):
            self = cls()
            self.client = client

            # Deal with 8 bit bytes for now
            if isinstance(source, str):
                source = source.encode('utf-8')

            # Convert MS-DOS or Mac line endings to the one true way
            source = source.replace(b'\r\n', b'\n').replace(b'\r', b'\n')
            sourcelist = splitter(source.decode())

            # Get the indentation at the start of the file
            indentation = '\n' + sourcelist[0]
            linenum = 1

            # Set up to iterate over the source and add to the destination list
            sourceiter = iter(sourcelist)
            next_ = sourceiter.__next__

            offset = 0 - len(next_())

            # Strip comment from first line
            if len(sourcelist) > 1 and sourcelist[1].startswith('#'):
                i = 1
                while len(sourcelist) > i and not sourcelist[i].startswith('\n'):
                    i += 1
                    offset -= len(next_())

            # Preallocate the list
            self.append(None)
            self *= len(sourcelist) // 2 + 1
            index = 0

            # Create all the tokens
            for token in sourceiter:
                whitespace = next_()
                t0 = token[0]
                if t0 not in delimiterset:
                    if t0 == '\n':
                        linenum += 1
                        indentation = token
                        offset -= len(token)
                        continue
                    else:
                        t0 = 'X'
                self[index] = (
                    offset,
                    t0,
                    token,
                    whitespace,
                    indentation,
                    linenum,
                    self,
                )
                index += 1
                offset -= len(token) + len(whitespace)

            # Add a sentinel
            self[index] = (offset, '@', '@', '', '', linenum + 1, self)
            self[index + 1 :] = []

            # Put everything we need in the actual object instantiation
            self.reverse()
            self.source = source
            self.next = self.pop
            self.push = self.append
            return self

        return newstring

    def peek(self):
        return self[-1]

    def lookahead(self, index=0):
        return self[-1 - index]

    @staticmethod
    def sourceloc(token):
        """Return the source location for a given token"""
        lineno = token[5]
        colno = offset = -token[0] + 1
        if lineno != 1:
            colno -= token[-1].source.rfind('\n', 0, offset) + 1
        return offset, lineno, colno

    @classmethod
    def error(cls, s, token):
        """error performs generic error reporting for tokens"""
        offset, lineno, colno = cls.sourceloc(token)

        if token[1] == '@':
            loc = 'at end of string'
        else:
            text = token[2]
            loc = 'line %s, column %s, text %s' % (lineno, colno, repr(text[:20]))

        err = RSONDecodeError('%s: %s' % (s, loc))
        err.pos = offset
        err.lineno = lineno
        err.colno = colno
        raise err


def make_hashable(what):
    try:
        hash(what)
        return what
    except TypeError:
        if isinstance(what, dict):
            return tuple(sorted(make_hashable(x) for x in what.items()))
        return tuple(make_hashable(x) for x in what)


class BaseObjects(object):

    # These hooks allow compatibility with simplejson
    object_hook = None
    object_pairs_hook = None
    array_hook = None

    # Stock object constructor does not cope with no keys
    disallow_missing_object_keys = True

    # Stock object constructor copes with multiple keys just fine
    disallow_multiple_object_keys = False

    # Default JSON requires string keys
    disallow_nonstring_keys = True

    class default_array(list):
        def __new__(self, startlist, token):
            return list(startlist)

    class default_object(dict):
        """By default, RSON objects are dictionaries that
        allow attribute access to their existing contents.
        """

        def __getattr__(self, key):
            return self[key]

        def __setattr__(self, key, value):
            self[key] = value

        def append(self, itemlist):
            mydict = self
            value = itemlist.pop()
            itemlist = [make_hashable(x) for x in itemlist]
            lastkey = itemlist.pop()

            if itemlist:
                itemlist.reverse()
                while itemlist:
                    key = itemlist.pop()
                    subdict = mydict.get(key)
                    if not isinstance(subdict, dict):
                        subdict = mydict[key] = type(self)()
                    mydict = subdict
            if isinstance(value, dict):
                oldvalue = mydict.get(lastkey)
                if isinstance(oldvalue, dict):
                    oldvalue.update(value)
                    return
            mydict[lastkey] = value

        def get_result(self, token):
            return self

    def object_type_factory(self):
        """This function returns constructors for RSON objects and arrays.
        It handles simplejson compatible hooks as well.
        """
        object_hook = self.object_hook
        object_pairs_hook = self.object_pairs_hook

        if object_pairs_hook is not None:

            class build_object(list):
                def get_result(self, token):
                    return object_pairs_hook([tuple(x) for x in self])

            self.disallow_multiple_object_keys = True
            self.disallow_nonstring_keys = True
        elif object_hook is not None:
            mydict = dict

            class build_object(list):
                def get_result(self, token):
                    return object_hook(mydict(self))

            self.disallow_multiple_object_keys = True
            self.disallow_nonstring_keys = True
        else:
            build_object = self.default_object

        build_array = self.array_hook or self.default_array
        return build_object, build_array


class Dispatcher(object):
    """Assumes that this is mixed-in to a class with an
    appropriate parser_factory() method.
    """

    @classmethod
    def dispatcher_factory(cls):

        self = cls()
        parser_factory = self.parser_factory
        parsercache = {}
        cached = parsercache.get
        default_loads = parser_factory()

        def loads(s, **kw):
            if not kw:
                return default_loads(s)

            key = tuple(sorted(kw.items()))
            func = cached(key)
            if func is None:
                # Begin some real ugliness here -- just modify our instance to
                # have the correct user variables for the initialization functions.
                # Seems to speed up simplejson testcases a bit.
                self.__dict__ = dict((x, y) for (x, y) in key if y is not None)
                func = parsercache[key] = parser_factory()

            return func(s)

        return loads


class QuotedToken(object):
    """Subclass or replace this if you don't like quoted string handling"""

    parse_encoded_chr = chr
    parse_quoted_str = staticmethod(lambda token, s, str=str: str(s))

    parse_join_str = u''.join
    cachestrings = False

    quoted_splitter = re.compile(r'(\\u[0-9a-fA-F]{4}|\\.|")').split
    quoted_mapper = {
        '\\\\': '\\',
        r'\"': '"',
        r'\/': '/',
        r'\b': '\b',
        r'\f': '\f',
        r'\n': '\n',
        r'\r': '\r',
        r'\t': '\t',
    }.get

    def quoted_parse_factory(self):
        quoted_splitter = self.quoted_splitter
        quoted_mapper = self.quoted_mapper
        parse_quoted_str = self.parse_quoted_str
        parse_encoded_chr = self.parse_encoded_chr
        parse_join_str = self.parse_join_str
        cachestrings = self.cachestrings
        triplequoted = self.triplequoted

        allow_double = sys.maxunicode > 65535

        def badstring(token, special):
            if token[2] != '"""' or triplequoted is None:
                token[-1].error(
                    'Invalid character in quoted string: %s' % repr(special), token
                )
            result = parse_quoted_str(token, triplequoted(token))
            if cachestrings:
                result = token[-1].stringcache(result, result)
            return result

        def parse(token, next_):
            s = token[2]
            if len(s) < 2 or not (s[0] == s[-1] == '"'):
                token[-1].error('No end quote on string', token)
            s = quoted_splitter(s[1:-1])
            result = parse_quoted_str(token, s[0])
            if len(s) > 1:
                result = [result]
                append = result.append
                s = iter(s)

                next_ = s.__next__
                next_()

                for special in s:
                    nonmatch = next_()
                    remap = quoted_mapper(special)
                    if remap is None:
                        if len(special) == 6:
                            uni = int(special[2:], 16)
                            if 0xD800 <= uni <= 0xDBFF and allow_double:
                                uni, nonmatch = parse_double_unicode(
                                    uni, nonmatch, next_, token
                                )
                            remap = parse_encoded_chr(uni)
                        else:
                            return badstring(token, special)
                    append(remap)
                    append(parse_quoted_str(token, nonmatch))

                result = parse_join_str(result)
            if cachestrings:
                result = token[-1].stringcache(result, result)
            return result

        def parse_double_unicode(uni, nonmatch, next_, token):
            """Munged version of UCS-4 code pair stuff from
            simplejson.
            """
            ok = True
            try:
                uni2 = next_()
                nonmatch2 = next_()
            except Exception:
                ok = False
            ok = ok and not nonmatch and uni2.startswith(b'\\u') and len(uni2) == 6
            if ok:
                nonmatch = uni2
                uni = 0x10000 + (((uni - 0xD800) << 10) | (int(uni2[2:], 16) - 0xDC00))
                return uni, nonmatch2
            token[-1].error(
                'Invalid second ch in double sequence: %s' % repr(nonmatch), token
            )

        return parse

    @staticmethod
    def triplequoted(token):
        tokens = token[-1]
        source = tokens.source
        result = []
        start = 3 - token[0]
        while 1:
            end = source.find('"""', start)
            if end < 0:
                tokens.error('Did not find end for triple-quoted string', token)
            if source[end - 1] != '\\':
                break
            result.append(source[start : end - 1])
            result.append('"""')
            start = end + 3
        result.append(source[start:end])
        offset = bisect.bisect(tokens, (-end - 2,))
        tokens[offset:] = []
        return ''.join(result)


class UnquotedToken(object):
    """Subclass or replace this if you don't like the unquoted
    token handling.  This is designed to be a superset of JSON:

      - Integers allowed to be expressed in octal, binary, or hex
        as well as decimal.

      - Integers can have embedded underscores.

      - Non-match of a special token will just be wrapped as a unicode
        string.

      - Numbers can be preceded by '+' as well s '-'
      - Numbers can be left-zero-filled
      - If a decimal point is present, digits are required on either side,
        but not both sides
    """

    use_decimal = False
    parse_int = staticmethod(lambda s: int(s.replace('_', ''), 0))
    parse_float = float
    parse_unquoted_str = staticmethod(lambda token: str(token[2]))

    special_strings = dict(true=True, false=False, null=None)

    unquoted_pattern = r'''
    (?:
        true | false | null       |     # Special JSON names
        (?P<num>
            [-+]?                       # Optional sign
            (?:
                0[xX](_*[0-9a-fA-F]+)+   | # Hex integer
                0[bB](_*[01]+)+          | # binary integer
                0[oO](_*[0-7]+)+         | # Octal integer
                \d+(_*\d+)*              | # Decimal integer
                (?P<float>
                    (?:
                      \d+(\.\d*)? |     # One or more digits,
                                        # optional frac
                      \.\d+             # Leading decimal point
                    )
                    (?:[eE][-+]?\d+)?   # Optional exponent
                )
            )
        )
    )  \Z                               # Match end of string
    '''

    def unquoted_parse_factory(self):
        unquoted_match = re.compile(self.unquoted_pattern, re.VERBOSE).match

        parse_unquoted_str = self.parse_unquoted_str
        parse_float = self.parse_float
        parse_int = self.parse_int
        special = self.special_strings

        if self.use_decimal:
            from decimal import Decimal

            parse_float = Decimal

        def parse(token, next_):
            s = token[2]
            m = unquoted_match(s)
            if m is None:
                return parse_unquoted_str(token)
            if m.group('num') is None:
                return special[s]
            if m.group('float') is None:
                return parse_int(s.replace('_', ''))
            return parse_float(s)

        return parse


class EqualToken(object):
    """Subclass or replace this if you don't like the = string handling"""

    encode_equals_str = None

    @staticmethod
    def parse_equals(stringlist, indent, token):
        """token probably not needed except maybe for error reporting.
        Replace this with something that does what you want.
        """
        # Strip any trailing whitespace to the right
        stringlist = [x.rstrip() for x in stringlist]

        # Strip any embedded comments
        stringlist = [x for x in stringlist if x.startswith(indent) or not x]

        # Strip trailing whitespace down below
        while stringlist and not stringlist[-1]:
            stringlist.pop()

        # Special cases for single line
        if not stringlist:
            return ''
        if len(stringlist) == 1:
            return stringlist[0].strip()

        # Strip whitespace on first line
        if stringlist and not stringlist[0]:
            stringlist.pop(0)

        # Dedent all the strings to one past the equals
        dedent = len(indent)
        stringlist = [x[dedent:] for x in stringlist]

        # Figure out if should dedent one more
        if min((not x and 500 or len(x) - len(x.lstrip())) for x in stringlist):
            stringlist = [x[1:] for x in stringlist]

        # Give every line its own linefeed (keeps later parsing from
        # treating this as a number, for example)
        stringlist.append('')

        # Return all joined up as a single unicode string
        return '\n'.join(stringlist)

    def equal_parse_factory(self, read_unquoted):

        parse_equals = self.parse_equals
        encoder = self.encode_equals_str

        if encoder is None:
            encoder = read_unquoted

        def parse(firsttok, next_):
            tokens = firsttok[-1]
            indent, linenum = firsttok[4:6]
            token = next_()
            while token[5] == linenum:
                token = next_()
            while token[4] > indent:
                token = next_()
            tokens.push(token)

            # Get rid of \n, and indent one past =
            indent = indent[1:] + ' '

            bigstring = tokens.source[-firsttok[0] + 1 : -token[0]]
            bigstring = bigstring.decode('utf-8')
            stringlist = bigstring.split('\n')
            stringlist[0] = indent + stringlist[0]
            token = list(firsttok)
            token[1:3] = '=', parse_equals(stringlist, indent, firsttok)
            return encoder(token, next)

        return parse


class RsonParser(object):
    """Parser for RSON"""

    disallow_trailing_commas = True
    disallow_rson_sublists = False
    disallow_rson_subdicts = False

    @staticmethod
    def post_parse(tokens, value):
        return value

    def client_info(self, parse_locals):
        pass

    def parser_factory(self):

        Tokenizer = self.Tokenizer
        tokenizer = Tokenizer.factory()
        error = Tokenizer.error

        read_unquoted = self.unquoted_parse_factory()
        read_quoted = self.quoted_parse_factory()
        parse_equals = self.equal_parse_factory(read_unquoted)
        new_object, new_array = self.object_type_factory()
        disallow_trailing_commas = self.disallow_trailing_commas
        disallow_missing_object_keys = self.disallow_missing_object_keys
        key_handling = [
            disallow_missing_object_keys,
            self.disallow_multiple_object_keys,
        ]
        disallow_nonstring_keys = self.disallow_nonstring_keys
        post_parse = self.post_parse

        def bad_array_element(token, next):
            error('Expected array element', token)

        def bad_dict_key(token, next):
            error('Expected dictionary key', token)

        def bad_dict_value(token, next):
            error('Expected dictionary value', token)

        def bad_top_value(token, next):
            error('Expected start of object', token)

        def bad_unindent(token, next):
            error('Unindent does not match any outer indentation level', token)

        def bad_indent(token, next):
            error('Unexpected indentation', token)

        def read_json_array(firsttok, next):
            result = new_array([], firsttok)
            append = result.append
            while 1:
                token = next()
                t0 = token[1]
                if t0 == ']':
                    if result and disallow_trailing_commas:
                        error('Unexpected trailing comma', token)
                    break
                append(json_value_dispatch(t0, bad_array_element)(token, next))
                delim = next()
                t0 = delim[1]
                if t0 == ',':
                    continue
                if t0 != ']':
                    if t0 == '@':
                        error('Unterminated list (no matching "]")', firsttok)
                    error('Expected "," or "]"', delim)
                break
            return result

        def read_json_dict(firsttok, next):
            result = new_object()
            append = result.append
            while 1:
                token = next()
                t0 = token[1]
                if t0 == '}':
                    if result and disallow_trailing_commas:
                        error('Unexpected trailing comma', token)
                    break
                key = json_value_dispatch(t0, bad_dict_key)(token, next)
                if disallow_nonstring_keys and not isinstance(key, str):
                    error('Non-string key %s not supported' % repr(key), token)
                token = next()
                t0 = token[1]
                if t0 != ':':
                    error('Expected ":" after dict key %s' % repr(key), token)
                token = next()
                t0 = token[1]
                value = json_value_dispatch(t0, bad_dict_value)(token, next)
                append([key, value])
                delim = next()
                t0 = delim[1]
                if t0 == ',':
                    continue
                if t0 != '}':
                    if t0 == '@':
                        error('Unterminated dict (no matching "}")', firsttok)
                    error('Expected "," or "}"', delim)
                break
            return result.get_result(firsttok)

        def read_rson_unquoted(firsttok, next):
            toklist = []
            linenum = firsttok[5]
            while 1:
                token = next()
                if token[5] != linenum or token[1] in ':=':
                    break
                toklist.append(token)
            firsttok[-1].push(token)
            if not toklist:
                return read_unquoted(firsttok, next)
            s = list(firsttok[2:4])
            for tok in toklist:
                s.extend(tok[2:4])
            result = list(firsttok)
            result[3] = s.pop()
            result[2] = ''.join(s)
            return read_unquoted(result, next)

        json_value_dispatch = {
            'X': read_unquoted,
            '[': read_json_array,
            '{': read_json_dict,
            '"': read_quoted,
        }.get

        rson_value_dispatch = {
            'X': read_rson_unquoted,
            '[': read_json_array,
            '{': read_json_dict,
            '"': read_quoted,
            '=': parse_equals,
        }

        if self.disallow_rson_sublists:
            rson_value_dispatch['['] = read_rson_unquoted

        if self.disallow_rson_subdicts:
            rson_value_dispatch['{'] = read_rson_unquoted

        rson_key_dispatch = rson_value_dispatch.copy()
        if disallow_missing_object_keys:
            del rson_key_dispatch['=']

        rson_value_dispatch = rson_value_dispatch.get
        rson_key_dispatch = rson_key_dispatch.get

        empty_object = new_object().get_result(None)
        empty_array = new_array([], None)
        empty_array_type = type(empty_array)
        empties = empty_object, empty_array

        def parse_recurse_array(stack, next, token, result):
            arrayindent, linenum = stack[-1][4:6]
            linenum -= not result
            while 1:
                thisindent, newlinenum = token[4:6]
                if thisindent != arrayindent:
                    if thisindent < arrayindent:
                        return result, token
                    if result:
                        stack.append(token)
                        lastitem = result[-1]
                        if lastitem == empty_array:
                            result[-1], token = parse_recurse_array(
                                stack, next, token, lastitem
                            )
                        elif lastitem == empty_object:
                            result[-1], token = parse_recurse_dict(
                                stack, next, token, lastitem
                            )
                        else:
                            result = None
                    if result:
                        stack.pop()
                        thisindent, newlinenum = token[4:6]
                        if thisindent <= arrayindent:
                            continue
                        bad_unindent(token, next)
                    bad_indent(token, next)
                if newlinenum <= linenum:
                    if token[1] in '=:':
                        error(
                            'Cannot mix list elements with dict (key/value) elements',
                            token,
                        )
                    error(
                        'Array elements must either be on separate lines or enclosed in []',
                        token,
                    )
                linenum = newlinenum
                value = rson_value_dispatch(token[1], bad_top_value)(token, next)
                result.append(value)
                token = next()

        def parse_one_dict_entry(stack, next, token, entry, mydict):
            arrayindent, linenum = stack[-1][4:6]
            while token[1] == ':':
                tok1 = next()
                thisindent, newlinenum = tok1[4:6]
                if newlinenum == linenum:
                    value = rson_value_dispatch(tok1[1], bad_top_value)(tok1, next)
                    token = next()
                    entry.append(value)
                    continue
                if thisindent <= arrayindent:
                    error('Expected indented line after ":"', token)
                token = tok1

            if not entry:
                error('Expected key', token)

            thisindent, newlinenum = token[4:6]
            if newlinenum == linenum and token[1] == '=':
                value = rson_value_dispatch(token[1], bad_top_value)(token, next)
                entry.append(value)
                token = next()
            elif thisindent > arrayindent:
                stack.append(token)
                value, token = parse_recurse(stack, next)
                if entry[-1] in empties:
                    if type(entry[-1]) is type(value):
                        entry[-1] = value
                    else:
                        error(
                            'Cannot load %s into %s' % (type(value), type(entry[-1])),
                            stack[-1],
                        )
                elif len(value) == 1 and type(value) is empty_array_type:
                    entry.extend(value)
                else:
                    entry.append(value)
                stack.pop()
            length = len(entry)
            if length != 2 and key_handling[length > 2]:
                if length < 2:
                    error('Expected ":" or "=", or indented line', token)
                error(
                    "rson client's object handlers do not support chained objects",
                    token,
                )
            if disallow_nonstring_keys:
                for key in entry[:-1]:
                    if not isinstance(key, str):
                        error('Non-string key %s not supported' % repr(key), token)
            mydict.append(entry)
            return token

        def parse_recurse_dict(stack, next, token, result):
            arrayindent = stack[-1][4]
            while 1:
                thisindent = token[4]
                if thisindent != arrayindent:
                    if thisindent < arrayindent:
                        return result.get_result(token), token
                    bad_unindent(token, next)
                key = rson_key_dispatch(token[1], bad_top_value)(token, next)
                stack[-1] = token
                token = parse_one_dict_entry(stack, next, next(), [key], result)

        def parse_recurse(stack, next, tokens=None):
            """parse_recurse ALWAYS returns a list or a dict.
            (or the user variants thereof)
            It is up to the caller to determine that it was an array
            of length 1 and strip the contents out of the array.
            """
            firsttok = stack[-1]
            value = rson_value_dispatch(firsttok[1], bad_top_value)(firsttok, next)

            # We return an array if the next value is on a new line and either
            # is at the same indentation, or the current value is an empty list

            token = next()
            if (
                token[5] != firsttok[5]
                and (token[4] <= firsttok[4] or value in empties)
                and disallow_missing_object_keys
            ):
                result = new_array([value], firsttok)
                if tokens is not None:
                    tokens.top_object = result
                return parse_recurse_array(stack, next, token, result)

            # Otherwise, return a dict
            result = new_object()
            if tokens is not None:
                tokens.top_object = result
            token = parse_one_dict_entry(stack, next, token, [value], result)
            return parse_recurse_dict(stack, next, token, result)

        def parse(source):
            tokens = tokenizer(source, None)
            tokens.stringcache = {}.setdefault
            tokens.client_info = client_info
            next = tokens.next
            value, token = parse_recurse([next()], next, tokens)
            if token[1] != '@':
                error('Unexpected additional data', token)

            # If it's a single item and we don't have a specialized
            # object builder, just strip the outer list.
            if (
                len(value) == 1
                and isinstance(value, list)
                and disallow_missing_object_keys
            ):
                value = value[0]
            return post_parse(tokens, value)

        client_info = self.client_info(locals())

        return parse


class RsonSystem(
    RsonParser, UnquotedToken, QuotedToken, EqualToken, Dispatcher, BaseObjects
):
    Tokenizer = Tokenizer


loads = RsonSystem.dispatcher_factory()



# -*- coding: utf-8 -*-

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['rst2pdf.pdfbuilder']

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'sphinx'
copyright = u'2009, RA'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = 'test'
# The full version, including alpha/beta/rc tags.
release = 'test'


# -- Options for PDF output ----------------------------------------------------

# Grouping the document tree into PDF files. List of tuples
# (source start file, target name, title, author).
pdf_documents = [
    (
        'index',
        u'index',
        u'test1',
        u'C.G.',
        {'pdf_use_index': True, 'pdf_use_modindex': True, 'pdf_use_coverpage': True},
    ),
    (
        'index2',
        u'index2',
        u'test2',
        u'C.G.',
        {
            'pdf_use_index': False,
            'pdf_use_modindex': False,
            'pdf_use_coverpage': False,
            'pdf_compressed': True,
        },
    ),
]
pdf_stylesheets = ['sphinx', 'sphinx-issue364']
pdf_use_index = False
pdf_use_modindex = False
pdf_use_coverpage = False
pdf_invariant = True
pdf_real_footnotes = True

# Set a consistent date for the cover page
today = 'April 29, 2018'


# -- General configuration -----------------------------------------------------
extensions = ['rst2pdf.pdfbuilder']
master_doc = 'index'

# -- Options for PDF output ----------------------------------------------------
# Grouping the document tree into PDF files (source start file, target file name, title, author).
pdf_documents = [('index', 'BreakLevel_2', 'Break Level Test', 'Rob Allen')]

pdf_stylesheets = ['sphinx']
pdf_use_toc = False
pdf_use_index = False
pdf_use_modindex = False
pdf_use_coverpage = False
pdf_invariant = True
pdf_breakside = 'any'

pdf_break_level = 2



"""
Copyright (c) 2020, Stephen Finucane <stephen@that.guru>

Automated testing for rst2pdf.

See LICENSE.txt for licensing terms
"""

import glob
import os
import shlex
import shutil
import subprocess
import tempfile

from packaging import version
import pytest

try:
    import fitz
except ImportError:
    import fitz_old as fitz


ROOT_DIR = os.path.realpath(os.path.dirname(__file__))
INPUT_DIR = os.path.join(ROOT_DIR, 'input')
OUTPUT_DIR = os.path.join(ROOT_DIR, 'output')
REFERENCE_DIR = os.path.join(ROOT_DIR, 'reference')


def can_run(command_list):
    def _can_run():
        try:
            p = subprocess.Popen(
                command_list,
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            p.terminate()
        except FileNotFoundError:
            return False
        return True

    return _can_run


check_dependency = {'plantuml': can_run(["plantuml", "-pipe"])}


def _get_metadata(pdf):
    metadata = pdf.metadata

    # Do not compare dates as they will differ each time the PDF is generated
    del metadata['creationDate']
    del metadata['modDate']

    # Do not compare the ReportLab producer string as it may differ between versions
    del metadata['producer']

    return metadata


def _get_pages(pdf):
    pages = []

    for page in pdf.pages():
        pages.append(page.get_text('blocks'))

    return pages


def compare_pdfs(path_a, path_b):
    try:
        pdf_a = fitz.open(path_a)
    except RuntimeError:
        pytest.fail(
            'Reference file at %r is not a valid PDF.'
            % (os.path.relpath(path_a, ROOT_DIR),),
            pytrace=False,
        )

    try:
        pdf_b = fitz.open(path_b)
    except RuntimeError:
        pytest.fail(
            'Output file at %r is not a valid PDF.'
            % (os.path.relpath(path_b, ROOT_DIR),),
            pytrace=False,
        )

    # sanity check

    assert pdf_a.is_pdf
    assert pdf_b.is_pdf

    # compare metadata

    assert _get_metadata(pdf_a) == _get_metadata(pdf_b)

    # compare content

    pages_a = _get_pages(pdf_a)
    pages_b = _get_pages(pdf_b)

    def fuzzy_coord_diff(coord_a, coord_b):
        diff = abs(coord_a - coord_b)
        threshold = 3  # 3 px is approximately 1.06mm
        assert (
            diff < threshold
        ), "Coordinates of the last printed block differs from the reference"

    def fuzzy_string_diff(string_a, string_b):
        a_is_image = string_a.startswith("<image: DeviceRGB")
        b_is_image = string_b.startswith("<image: DeviceRGB")
        if a_is_image and b_is_image:
            # We can't necessarily control the image metadata text in the block (e.g. from plantuml), so we do not
            # check it.
            return

        words_a = string_a.split()
        words_b = string_b.split()
        assert (
            words_a == words_b
        ), "Text of the last printed block differs from the reference"

    assert len(pages_a) == len(pages_b), "Number of pages differs from the reference"
    page_no = 0
    for page_a, page_b in zip(pages_a, pages_b):
        page_no = page_no + 1
        print(f"++ Page {page_no} ++")
        print(f"page_a: {page_a}")
        print(f"page_b: {page_b}")
        print("number of blocks in page_a: %s" % len(page_a))
        print("number of blocks in page_b: %s" % len(page_b))
        assert len(page_a) == len(
            page_b
        ), f"Number of blocks on page {page_no} differs from the reference"
        for block_a, block_b in zip(page_a, page_b):
            # each block has the following format:
            #
            # (x0, y0, x1, y1, "lines in block", block_type, block_no)
            #
            # block_type and block_no should remain unchanged, but it's
            # possible for the blocks to move around the document slightly and
            # the text refold without breaking entirely
            print(f"block_a: {block_a}")
            print(f"block_b: {block_b}")
            fuzzy_coord_diff(block_a[0], block_b[0])
            fuzzy_coord_diff(block_a[1], block_b[1])
            fuzzy_coord_diff(block_a[2], block_b[2])
            fuzzy_coord_diff(block_a[3], block_b[3])
            fuzzy_string_diff(block_a[4], block_b[4])
            assert block_a[5] == block_b[5]
            assert block_a[6] == block_b[6]


class File(pytest.File):

    if version.parse(pytest.__version__) < version.parse('5.4.0'):

        @classmethod
        def from_parent(cls, parent, path):
            return cls(parent=parent, path=path)


class RstFile(File):
    def collect(self):
        name = os.path.splitext(self.fspath.basename)[0]
        yield RstItem.from_parent(parent=self, name=name)


class SphinxFile(File):
    def collect(self):
        name = os.path.split(self.fspath.dirname)[-1]
        yield SphinxItem.from_parent(parent=self, name=name)


class Item(pytest.Item):

    if version.parse(pytest.__version__) < version.parse('5.4.0'):

        @classmethod
        def from_parent(cls, parent, name):
            return cls(parent=parent, name=name)

    def _build(self):
        raise NotImplementedError

    def _fail(self, msg, output=None):
        pytest.fail(
            f'{msg}: \n\n{output.decode("utf-8")}' if output else msg,
            pytrace=False,
        )

    def runtest(self):
        __tracebackhide__ = True

        # if '.ignore' file present, skip test

        ignore_file = os.path.join(INPUT_DIR, self.name + '.ignore')
        if os.path.exists(ignore_file):
            with open(ignore_file) as fh:
                ignore_reason = fh.read()

            pytest.skip(ignore_reason)

        # if '.depends' file is present, check if all dependencies are
        # satisfied, otherwise skip test

        depends_file = os.path.join(INPUT_DIR, self.name + '.depends')
        if os.path.exists(depends_file):
            with open(depends_file) as fh:
                for line in fh:
                    dep = line.rstrip('\n')
                    try:
                        if not check_dependency[dep]():
                            pytest.skip("Unmet dependency for test: %s" % line)
                    except KeyError:
                        pytest.skip("Unknown test dependency: %s" % line)

        # run the actual test

        retcode, output = self._build()

        # verify results

        retcode_file = os.path.join(INPUT_DIR, self.name + '.retcode')
        if os.path.exists(retcode_file):
            with open(retcode_file) as f:
                first_line = f.readline()
                expected_retcode = int(first_line)
                if expected_retcode == retcode:
                    return
                else:
                    self._fail(
                        'Exit code of %d did not match expected %d'
                        % (retcode, expected_retcode),
                        output,
                    )
        elif retcode > 0:
            self._fail('Call failed with %d' % retcode, output)

        no_pdf = os.path.exists(os.path.join(INPUT_DIR, self.name + '.nopdf'))
        if no_pdf:
            return

        output_file = os.path.join(OUTPUT_DIR, self.name + '.pdf')
        reference_file = os.path.join(REFERENCE_DIR, self.name + '.pdf')

        if not os.path.exists(reference_file):
            self._fail(
                'No reference file at %r to compare against.'
                % (os.path.relpath(output_file, ROOT_DIR),),
            )

        if os.path.isdir(output_file):
            if not os.path.isdir(reference_file):
                self._fail(
                    'Mismatch between type of output (directory) and '
                    'reference (file)',
                    output,
                )
            output_files = glob.glob(os.path.join(output_file, '*.pdf'))
            reference_files = glob.glob(os.path.join(reference_file, '*.pdf'))
        else:
            if not os.path.isfile(reference_file):
                self._fail(
                    'Mismatch between type of output (file) and reference '
                    '(directory)',
                    output,
                )
            output_files = [output_file]
            reference_files = [reference_file]

        if len(reference_files) != len(output_files):
            self._fail(
                'Mismatch between number of files expected and generated',
                output,
            )

        reference_files.sort()
        output_files.sort()

        for ref_pdf, out_pdf in zip(reference_files, output_files):
            try:
                compare_pdfs(ref_pdf, out_pdf)
            except AssertionError as exc:
                raise CompareException(exc)

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""
        if isinstance(excinfo.value, CompareException):
            return excinfo.exconly()

        return super(Item, self).repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, 0, self.name


class RstItem(Item):
    def _build(self):
        input_ref = self.name + '.rst'
        output_pdf = os.path.join(OUTPUT_DIR, self.name + '.pdf')
        output_log = os.path.join(OUTPUT_DIR, self.name + '.log')

        for path in (output_log, output_pdf):
            if os.path.exists(path):
                if os.path.isdir(path):
                    shutil.rmtree(path)
                else:
                    os.remove(path)

        cmd = ['rst2pdf', '--lang', 'en', '--date-invariant', '-v', input_ref]

        cli_file = os.path.join(INPUT_DIR, self.name + '.cli')
        if os.path.exists(cli_file):
            with open(cli_file) as fh:
                cmd += shlex.split(fh.read())

        style_file = os.path.join(INPUT_DIR, self.name + '.yaml')
        if os.path.exists(style_file):
            cmd += ['-s', os.path.basename(style_file)]

        if '-o' not in cmd:
            cmd += ['-o', output_pdf]

        try:
            output = subprocess.check_output(
                cmd,
                cwd=INPUT_DIR,
                stderr=subprocess.STDOUT,
            )
            retcode = 0
        except subprocess.CalledProcessError as exc:
            output = exc.output
            retcode = exc.returncode

        with open(output_log, 'wb') as fh:
            fh.write(output)

        expected_log_file = os.path.join(
            os.path.join(INPUT_DIR, self.name + '.expected_log')
        )
        output_file = os.path.join(OUTPUT_DIR, self.name + '.pdf')
        no_pdf = os.path.exists(os.path.join(INPUT_DIR, self.name + '.nopdf'))

        if os.path.exists(expected_log_file):
            # look for each line in expected_log_file within the output log
            output_string = output.decode("utf-8")
            with open(expected_log_file) as expected_log_file:
                for message in expected_log_file:
                    message = message.strip()
                    if message and message not in output_string:
                        self._fail(
                            'Log message %s was not found' % message,
                            output,
                        )
            if no_pdf:
                # As we are looking for a log entry, we ignore any created pdf file if a .nopdf
                # file exists. That is, it doesn't matter if the file was created or not.
                return retcode, output

        if not os.path.exists(output_file) and not no_pdf:
            self._fail(
                'File %r was not generated' % (os.path.relpath(output_file, ROOT_DIR),),
                output,
            )
        elif os.path.exists(output_file) and no_pdf:
            self._fail(
                'File %r was erroneously generated'
                % (os.path.relpath(output_file, ROOT_DIR),),
                output,
            )

        return retcode, output


class SphinxItem(Item):
    def _build(self):
        __tracebackhide__ = True

        output_pdf = os.path.join(OUTPUT_DIR, self.name + '.pdf')
        output_log = os.path.join(OUTPUT_DIR, self.name + '.log')

        for path in (output_log, output_pdf):
            if os.path.exists(path):
                if os.path.isdir(path):
                    shutil.rmtree(path)
                else:
                    os.remove(path)

        input_dir = os.path.join(INPUT_DIR, self.name)
        build_dir = tempfile.mkdtemp(prefix='rst2pdf-sphinx-')

        cmd = ['sphinx-build', '-b', 'pdf', input_dir, build_dir]

        try:
            output = subprocess.check_output(
                cmd,
                cwd=INPUT_DIR,
                stderr=subprocess.STDOUT,
            )
            retcode = 0
        except subprocess.CalledProcessError as exc:
            output = exc.output
            retcode = exc.returncode

        with open(output_log, 'wb') as fh:
            fh.write(output)

        pdf_files = glob.glob(os.path.join(build_dir, '*.pdf'))
        if pdf_files:
            if len(pdf_files) == 1:
                shutil.copyfile(pdf_files[0], output_pdf)
            else:
                shutil.copytree(build_dir, output_pdf)
        else:
            self._fail('Output PDF not generated', output)

        shutil.rmtree(build_dir)

        return retcode, output


class CompareException(Exception):
    """Custom exception for error reporting."""


def pytest_collect_file(file_path, parent):

    parent_directory = file_path.parents[0]
    if "input" not in str(parent_directory):
        return

    parent_dir = os.path.split(parent_directory)[-1]

    if file_path.suffix == '.rst' and parent_dir == 'input':
        return RstFile.from_parent(parent=parent, path=file_path)
    elif file_path.name == 'conf.py' and parent_dir.startswith('sphinx'):
        return SphinxFile.from_parent(parent=parent, path=file_path)


collect_ignore = ['tests/input/*.py']



# -*- coding: utf-8 -*-

# :Author: a Pygments author|contributor; Felix Wiemann; Guenter Milde
# :Date: $Date$
# :Copyright: This module has been placed in the public domain.
#
# This is a merge of `Using Pygments in ReST documents`_ from the pygments_
# documentation, and a `proof of concept`_ by Felix Wiemann.
#
# ========== ===========================================================
# 2007-06-01 Removed redundancy from class values.
# 2007-06-04 Merge of successive tokens of same type
#            (code taken from pygments.formatters.others).
# 2007-06-05 Separate docutils formatter script
#            Use pygments' CSS class names (like the html formatter)
#            allowing the use of pygments-produced style sheets.
# 2007-06-07 Merge in the formatting of the parsed tokens
#            (misnamed as docutils_formatter) as class DocutilsInterface
# 2007-06-08 Failsave implementation (fallback to a standard literal block
#            if pygments not found)
# ========== ===========================================================
#
# ::

"""Define and register a code-block directive using pygments"""


# Requirements
# ------------
# ::

import codecs

from docutils import nodes
from docutils.parsers.rst import directives

try:
    import pygments
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters.html import _get_ttype_class
except ImportError:
    pass

from ..log import log


# Customisation
# -------------
#
# Do not insert inline nodes for the following tokens.
# (You could add e.g. Token.Punctuation like ``['', 'p']``.) ::

unstyled_tokens = ['']


# DocutilsInterface
# -----------------
#
# This interface class combines code from
# pygments.formatters.html and pygments.formatters.others.
#
# It does not require anything of docutils and could also become a part of
# pygments::


class DocutilsInterface(object):
    """Parse `code` string and yield "classified" tokens.

    Arguments

      code     -- string of source code to parse
      language -- formal language the code is written in.

    Merge subsequent tokens of the same token-type.

    Yields the tokens as ``(ttype_class, value)`` tuples,
    where ttype_class is taken from pygments.token.STANDARD_TYPES and
    corresponds to the class argument used in pygments html output.

    """

    def __init__(self, code, language, custom_args={}):
        self.code = code
        self.language = language
        self.custom_args = custom_args

    def lex(self):
        # Get lexer for language (use text as fallback)
        try:
            if self.language and str(self.language).lower() != 'none':
                lexer = get_lexer_by_name(self.language.lower(), **self.custom_args)
            else:
                lexer = get_lexer_by_name('text', **self.custom_args)
        except ValueError:
            log.info("no pygments lexer for %s, using 'text'" % self.language)
            # what happens if pygment isn't present ?
            lexer = get_lexer_by_name('text')
        return pygments.lex(self.code, lexer)

    def join(self, tokens):
        """join subsequent tokens of same token-type"""
        tokens = iter(tokens)
        (lasttype, lastval) = next(tokens)
        for ttype, value in tokens:
            if ttype is lasttype:
                lastval += value
            else:
                yield (lasttype, lastval)
                (lasttype, lastval) = (ttype, value)
        yield (lasttype, lastval)

    def __iter__(self):
        """parse code string and yield "clasified" tokens"""
        try:
            tokens = self.lex()
        except IOError:
            log.info("Pygments lexer not found, using fallback")
            # TODO: write message to INFO
            yield ('', self.code)
            return

        for ttype, value in self.join(tokens):
            yield (_get_ttype_class(ttype), value)


# code_block_directive
# --------------------
# ::


def code_block_directive(
    name,
    arguments,
    options,
    content,
    lineno,
    content_offset,
    block_text,
    state,
    state_machine,
):
    """Parse and classify content of a code_block."""
    if 'include' in options:
        try:
            if 'encoding' in options:
                encoding = options['encoding']
            else:
                encoding = 'utf-8'
            content = codecs.open(options['include'], 'r', encoding).read().rstrip()
        except (IOError, UnicodeError):  # no file or problem finding it or reading it
            log.error('Error reading file: "%s" L %s' % (options['include'], lineno))
            content = u''
        line_offset = 0
        if content:
            # here we define the start-at and end-at options
            # so that limit is included in extraction
            # this is different than the start-after directive of docutils
            # (docutils/parsers/rst/directives/misc.py L73+)
            # which excludes the beginning
            # the reason is we want to be able to define a start-at like
            # def mymethod(self)
            # and have such a definition included

            after_text = options.get('start-at', None)
            if after_text:
                # skip content in include_text before *and NOT incl.* a matching text
                after_index = content.find(after_text)
                if after_index < 0:
                    raise state_machine.reporter.severe(
                        'Problem with "start-at" option of "%s" '
                        'code-block directive:\nText not found.' % options['start-at']
                    )
                # patch mmueller start
                # Move the after_index to the beginning of the line with the
                # match.
                for char in content[after_index:0:-1]:
                    # codecs always opens binary. This works with '\n', '\r' and
                    # '\r\n'. We are going backwards, so '\n' is found first
                    # in '\r\n'.
                    # Going with .splitlines() seems more appropriate
                    # but needs a few more changes.
                    if char == u'\n' or char == u'\r':
                        break
                    after_index -= 1
                # patch mmueller end

                content = content[after_index:]
                line_offset = len(content[:after_index].splitlines()) - 1

            after_text = options.get('start-after', None)
            if after_text:
                # skip content in include_text before *and incl.* a matching text
                after_index = content.find(after_text)
                if after_index < 0:
                    raise state_machine.reporter.severe(
                        'Problem with "start-after" option of "%s" '
                        'code-block directive:\nText not found.'
                        % options['start-after']
                    )
                after_index = after_index + len(after_text)

                # Move the after_index to the start of the line after the match
                for char in content[after_index:]:
                    if char == u'\n':
                        break
                    after_index += 1

                line_offset = len(content[:after_index].splitlines())
                content = content[after_index:]

            # same changes here for the same reason
            before_text = options.get('end-at', None)
            if before_text:
                # skip content in include_text after *and incl.* a matching text
                before_index = content.find(before_text)
                if before_index < 0:
                    raise state_machine.reporter.severe(
                        'Problem with "end-at" option of "%s" '
                        'code-block directive:\nText not found.' % options['end-at']
                    )
                content = content[: before_index + len(before_text)]

            before_text = options.get('end-before', None)
            if before_text:
                # skip content in include_text after *and NOT incl.* a matching text
                before_index = content.find(before_text)
                if before_index < 0:
                    raise state_machine.reporter.severe(
                        'Problem with "end-before" option of "%s" '
                        'code-block directive:\nText not found.' % options['end-before']
                    )
                content = content[:before_index]

    else:
        line_offset = options.get('linenos_offset')
        content = u'\n'.join(content)

    if 'tabsize' in options:
        tabw = options['tabsize']
    else:
        tabw = int(options.get('tab-width', 8))

    content = content.replace('\t', ' ' * tabw)

    # hl_lines is the option, but if it isn't used, the alias emphasize-lines is also supported for Sphinx compatibility
    hl_lines = options.get('hl_lines', [])
    # if hl_lines isn't used, check if emphasize-lines should be
    if hl_lines == []:
        hl_lines = options.get('emphasize-lines', [])

    # Here the value of "linenos" will generally be the empty string if the user supplied and desired it, but we also
    # want to allow it to be present and explicitly disabled
    withln = options.get("linenos", False) not in ("false", "False", False)
    if 'linenos_offset' not in options:
        line_offset = 0

    try:
        language = arguments[0]
    except IndexError:
        language = 'text'

    # create a literal block element and set class argument
    code_block = nodes.literal_block(classes=["code", language])

    lineno = 1 + line_offset
    total_lines = content.count('\n') + 1 + line_offset
    if withln:
        lnwidth = len(str(total_lines))
        fstr = "\n%%%dd " % lnwidth
        linenumber_cls = 'linenumber'
        if hl_lines and lineno not in hl_lines:
            linenumber_cls = 'pygments-diml'
        code_block += nodes.inline(
            fstr[1:] % lineno, fstr[1:] % lineno, classes=[linenumber_cls]
        )

    # parse content with pygments and add to code_block element
    for cls, value in DocutilsInterface(content, language, options):
        if hl_lines and lineno not in hl_lines:
            cls = "diml"
        if withln and "\n" in value:
            # Split on the "\n"s
            values = value.split("\n")
            # The first piece, pass as-is
            c = ''
            if cls != '':
                c = 'pygments-diml'
            code_block += nodes.inline(values[0], values[0], classes=[c])

            # On the second and later pieces, insert \n and linenos
            linenos = range(lineno, lineno + len(values))
            for chunk, ln in list(zip(values, linenos))[1:]:
                if ln <= total_lines:
                    linenumber_cls = 'linenumber'
                    c = ''
                    if hl_lines and (ln) not in hl_lines:
                        linenumber_cls = 'pygments-diml'
                        c = 'pygments-diml'

                    code_block += nodes.inline(
                        fstr % ln, fstr % ln, classes=[linenumber_cls]
                    )
                    code_block += nodes.inline(chunk, chunk, classes=[c])
            lineno += len(values) - 1

        elif cls in unstyled_tokens:
            if "\n" in value:
                lineno = lineno + value.count("\n")
            # insert as Text to decrease the verbosity of the output.
            code_block += nodes.Text(value, value)
        else:
            if "\n" in value:
                lineno = lineno + value.count("\n")
            code_block += nodes.inline(value, value, classes=["pygments-" + cls])

    return [code_block]


# Custom argument validators
# --------------------------
# ::
#
# Move to separated module??


def zero_or_positive_int(argument):
    """
    Converts a string into python positive integer including zero.
    None is a special case; it is regarded as zero.
    """
    if argument is None:
        return 0
    elif argument == '0':
        return 0
    else:
        return directives.positive_int(argument)


def string_list(argument):
    """
    Converts a space- or comma-separated list of values into a python list
    of strings.
    (Directive option conversion function)
    Based in positive_int_list of docutils.parsers.rst.directives
    """
    if ',' in argument:
        entries = argument.split(',')
    else:
        entries = argument.split()
    return entries


def string_bool(argument):
    """
    Converts True, true, False, False in python boolean values
    """
    if argument is None:
        msg = 'argument required but none supplied; choose from "True" or "False"'
        raise ValueError(msg)

    elif argument.lower() == 'true':
        return True
    elif argument.lower() == 'false':
        return False
    else:
        raise ValueError('"%s" unknown; choose from "True" or "False"' % argument)


def csharp_unicodelevel(argument):
    return directives.choice(argument, ('none', 'basic', 'full'))


def lhs_litstyle(argument):
    return directives.choice(argument, ('bird', 'latex'))


def raw_compress(argument):
    return directives.choice(argument, ('gz', 'bz2'))


# Register Directive
# ------------------
# ::

code_block_directive.arguments = (0, 1, 1)
code_block_directive.content = 1
code_block_directive.options = {
    'include': directives.unchanged_required,
    'start-at': directives.unchanged_required,
    'end-at': directives.unchanged_required,
    'start-after': directives.unchanged_required,
    'end-before': directives.unchanged_required,
    'linenos': directives.unchanged,
    'linenos_offset': zero_or_positive_int,
    'tab-width': directives.unchanged,
    'hl_lines': directives.positive_int_list,
    'emphasize-lines': directives.positive_int_list,
    # generic
    'stripnl': string_bool,
    'stripall': string_bool,
    'ensurenl': string_bool,
    'tabsize': directives.positive_int,
    'encoding': directives.encoding,
    # Lua
    'func_name_hightlighting': string_bool,
    'disabled_modules': string_list,
    # Python Console
    'python3': string_bool,
    # Delphi
    'turbopascal': string_bool,
    'delphi': string_bool,
    'freepascal': string_bool,
    'units': string_list,
    # Modula2
    'pim': string_bool,
    'iso': string_bool,
    'objm2': string_bool,
    'gm2ext': string_bool,
    # CSharp
    'unicodelevel': csharp_unicodelevel,
    # Literate haskell
    'litstyle': lhs_litstyle,
    # Raw
    'compress': raw_compress,
    # Rst
    'handlecodeblocks': string_bool,
    # Php
    'startinline': string_bool,
    'funcnamehighlighting': string_bool,
    'disabledmodules': string_list,
}


# .. _doctutils: http://docutils.sf.net/
# .. _pygments: http://pygments.org/
# .. _Using Pygments in ReST documents: http://pygments.org/docs/rstdirective/
# .. _proof of concept:
#      http://article.gmane.org/gmane.text.docutils.user/3689
#
# Test output
# -----------
#
# If called from the command line, call the docutils publisher to render the
# input::

if __name__ == '__main__':
    from docutils.core import publish_cmdline, default_description
    from docutils.parsers.rst import directives

    directives.register_directive('code-block', code_block_directive)
    description = "code-block directive test output" + default_description
    try:
        import locale

        locale.setlocale(locale.LC_ALL, '')
    except Exception:
        pass
    publish_cmdline(writer_name='html', description=description)



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

from copy import copy
import os
import os.path
import sys
import re

import docutils.nodes
import reportlab
import reportlab.lib.colors as colors
from reportlab.lib.fonts import addMapping
import reportlab.lib.pagesizes as pagesizes
from reportlab.lib.styles import StyleSheet1, ParagraphStyle
import reportlab.lib.units as units
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.pdfbase import pdfmetrics
import reportlab.rl_config
import yaml

from . import findfonts
from .log import log
from .rson import loads as rson_loads

unit_separator = re.compile('(-?[0-9.]*)')


class StyleSheet(object):
    """Class to handle a collection of stylesheets"""

    @staticmethod
    def stylepairs(data):
        """Allows pairs of style information to be expressed
        in canonical reportlab list of two-item list/tuple,
        or in a more human-readable dictionary.
        """
        styles = data.get('styles', {})
        try:
            stylenames = list(styles.keys())
        except AttributeError:
            for style in styles:
                yield style
            return

        # Traditional reportlab styles are in ordered (key, value)
        # tuples.  We also support dictionary lookup.  This is not
        # necessarily ordered.

        # The only problem with dictionary lookup is that
        # we need to insure that parents are processed before
        # their children.  This loop is a little ugly, but
        # gets the job done.

        while stylenames:
            name = stylenames.pop()
            parent = styles[name].get('parent')
            if parent not in stylenames:
                yield name, styles[name]
                continue
            names = [name]
            while parent in stylenames:
                stylenames.remove(parent)
                names.append(parent)
                parent = styles[names[-1]].get('parent')
            while names:
                name = names.pop()
                yield name, styles[name]

    def __init__(
        self,
        flist,
        font_path=None,
        style_path=None,
        def_dpi=300,
        record_dependencies=None,
    ):
        log.info('Using stylesheets: %s' % ','.join(flist))

        self.suppress_undefined_style_warning = False

        # find base path
        if hasattr(sys, 'frozen'):
            self.PATH = os.path.abspath(os.path.dirname(sys.executable))
        else:
            self.PATH = os.path.abspath(os.path.dirname(__file__))

        # flist is a list of stylesheet filenames.
        # They will be loaded and merged in order.
        # but the two default stylesheets will always
        # be loaded first
        flist = [
            os.path.join(self.PATH, 'styles', 'styles.yaml'),
            os.path.join(self.PATH, 'styles', 'default.yaml'),
        ] + flist

        self.def_dpi = def_dpi
        if font_path is None:
            font_path = []
        font_path += ['.', os.path.join(self.PATH, 'fonts')]
        self.FontSearchPath = list(map(os.path.expanduser, font_path))

        if style_path is None:
            style_path = []
        style_path += [
            '.',
            os.path.join(self.PATH, 'styles'),
            '~/.rst2pdf/styles',
        ]
        self.StyleSearchPath = list(map(os.path.expanduser, style_path))

        # Remove duplicates but preserve order. Not very efficient, but these are short lists
        self.FontSearchPath = [
            x
            for (i, x) in enumerate(self.FontSearchPath)
            if self.FontSearchPath.index(x) == i
        ]
        self.StyleSearchPath = [
            x
            for (i, x) in enumerate(self.StyleSearchPath)
            if self.StyleSearchPath.index(x) == i
        ]

        log.info('FontPath:%s' % self.FontSearchPath)
        log.info('StylePath:%s' % self.StyleSearchPath)

        findfonts.flist = self.FontSearchPath
        # Page width, height
        self.pw = 0
        self.ph = 0

        # Page size [w,h]
        self.ps = None

        # Margins (top,bottom,left,right,gutter)
        self.tm = 0
        self.bm = 0
        self.lm = 0
        self.rm = 0
        self.gm = 0

        # text width
        self.tw = 0

        # Default emsize, later it will be the fontSize of the base style
        self.emsize = 10

        self.languages = []
        self.record_dependencies = record_dependencies

        ssdata = self.readSheets(flist)

        # Get pageSetup data from all stylessheets in order:
        self.ps = pagesizes.A4
        self.page = {}
        for data, ssname in ssdata:
            page = data.get('pageSetup', {})
            if page:
                self.page.update(page)
                pgs = page.get('size', None)
                if pgs:  # A standard size
                    pgs = pgs.upper()
                    if pgs in pagesizes.__dict__:
                        self.ps = list(pagesizes.__dict__[pgs])
                        self.psname = pgs
                        if 'width' in self.page:
                            del self.page['width']
                        if 'height' in self.page:
                            del self.page['height']
                    elif pgs.endswith('-LANDSCAPE'):
                        self.psname = pgs.split('-')[0]
                        self.ps = list(
                            pagesizes.landscape(pagesizes.__dict__[self.psname])
                        )
                        if 'width' in self.page:
                            del self.page['width']
                        if 'height' in self.page:
                            del self.page['height']
                    else:
                        log.critical(
                            'Unknown page size %s in stylesheet %s'
                            % (page['size'], ssname)
                        )
                        continue
                else:  # A custom size
                    if 'size' in self.page:
                        del self.page['size']
                    # The sizes are expressed in some unit.
                    # For example, 2cm is 2 centimeters, and we need
                    # to do 2*cm (cm comes from reportlab.lib.units)
                    if 'width' in page:
                        self.ps[0] = self.adjustUnits(page['width'])
                    if 'height' in page:
                        self.ps[1] = self.adjustUnits(page['height'])
                self.pw, self.ph = self.ps
                if 'margin-left' in page:
                    self.lm = self.adjustUnits(page['margin-left'])
                if 'margin-right' in page:
                    self.rm = self.adjustUnits(page['margin-right'])
                if 'margin-top' in page:
                    self.tm = self.adjustUnits(page['margin-top'])
                if 'margin-bottom' in page:
                    self.bm = self.adjustUnits(page['margin-bottom'])
                if 'margin-gutter' in page:
                    self.gm = self.adjustUnits(page['margin-gutter'])
                if 'spacing-header' in page:
                    self.ts = self.adjustUnits(page['spacing-header'])
                if 'spacing-footer' in page:
                    self.bs = self.adjustUnits(page['spacing-footer'])
                if 'firstTemplate' in page:
                    self.firstTemplate = page['firstTemplate']

                # tw is the text width.
                # We need it to calculate header-footer height
                # and compress literal blocks.
                self.tw = self.pw - self.lm - self.rm - self.gm

        # Get page templates from all stylesheets
        self.pageTemplates = {}
        for data, ssname in ssdata:
            templates = data.get('pageTemplates', {})
            # templates is a dictionary of pageTemplates
            for key in templates:
                template = templates[key]
                # template is a dict.
                # template[´frames'] is a list of frames
                if key in self.pageTemplates:
                    self.pageTemplates[key].update(template)
                else:
                    self.pageTemplates[key] = template

        # Get font aliases from all stylesheets in order
        self.fontsAlias = {}
        for data, ssname in ssdata:
            self.fontsAlias.update(data.get('fontsAlias', {}))

        embedded_fontnames = []
        self.embedded = []
        # Embed all fonts indicated in all stylesheets
        for data, ssname in ssdata:
            embedded = data.get('embeddedFonts', [])

            for font in embedded:
                try:
                    # (removed the feature that supported a string here, this isn't
                    # documented or supported)

                    # Each "font" is a list of four files, which will be
                    # used for regular / bold / italic / bold+italic
                    # versions of the font.
                    # If your font doesn't have one of them, just repeat
                    # the regular font.

                    # Example, using the Tuffy font from
                    # http://tulrich.com/fonts/
                    # "embeddedFonts" : [
                    #                    ["Tuffy.ttf",
                    #                     "Tuffy_Bold.ttf",
                    #                     "Tuffy_Italic.ttf",
                    #                     "Tuffy_Bold_Italic.ttf"]
                    #                   ],

                    # The fonts will be registered with the file name,
                    # minus the extension.

                    if font[0].lower().endswith('.ttf'):  # A True Type font
                        for variant in font:
                            location = self.findFont(variant)
                            # strip extension and leading path to get the name
                            # to register the font under
                            filename = os.path.basename(variant)
                            if self.record_dependencies:
                                self.record_dependencies.add(filename)
                            fontname = str(filename.split('.')[0])
                            pdfmetrics.registerFont(TTFont(fontname, location))
                            log.info(
                                'Registering font: %s from %s' % (fontname, location)
                            )
                            self.embedded.append(fontname)

                        # And map them all together
                        regular, bold, italic, bolditalic = [
                            variant.split('.')[0] for variant in font
                        ]
                        addMapping(regular, 0, 0, regular)
                        addMapping(regular, 0, 1, italic)
                        addMapping(regular, 1, 0, bold)
                        addMapping(regular, 1, 1, bolditalic)
                    else:  # A Type 1 font
                        # For type 1 fonts we require
                        # [FontName,regular,italic,bold,bolditalic]
                        # where each variant is a (pfbfile,afmfile) pair.
                        # For example, for the URW palladio from TeX:
                        # ["Palatino",("uplr8a.pfb","uplr8a.afm"),
                        #             ("uplri8a.pfb","uplri8a.afm"),
                        #             ("uplb8a.pfb","uplb8a.afm"),
                        #             ("uplbi8a.pfb","uplbi8a.afm")]
                        regular = pdfmetrics.EmbeddedType1Face(*font[1])
                        italic = pdfmetrics.EmbeddedType1Face(*font[2])
                        bold = pdfmetrics.EmbeddedType1Face(*font[3])
                        bolditalic = pdfmetrics.EmbeddedType1Face(*font[4])

                except Exception as e:
                    try:
                        if isinstance(font, list):
                            fname = font[0]
                        else:
                            fname = font
                        log.critical(
                            "Error processing font %s: %s",
                            os.path.splitext(fname)[0],
                            str(e),
                        )
                        sys.exit(1)
                    except Exception as e:
                        log.critical("Error processing font %s: %s", fname, str(e))
                        sys.exit(1)

        # Go though all styles in all stylesheets and find all fontNames.
        # Then decide what to do with them
        for data, ssname in ssdata:
            for [skey, style] in self.stylepairs(data):
                for key in style:
                    if key == 'fontName' or key.endswith('FontName'):
                        # It's an alias, replace it
                        if style[key] in self.fontsAlias:
                            style[key] = self.fontsAlias[style[key]]
                        # Embedded already, nothing to do
                        if style[key] in self.embedded:
                            continue
                        # Standard font, nothing to do
                        if style[key] in (
                            "Courier",
                            "Courier-Bold",
                            "Courier-BoldOblique",
                            "Courier-Oblique",
                            "Helvetica",
                            "Helvetica-Bold",
                            "Helvetica-BoldOblique",
                            "Helvetica-Oblique",
                            "Symbol",
                            "Times-Bold",
                            "Times-BoldItalic",
                            "Times-Italic",
                            "Times-Roman",
                            "ZapfDingbats",
                        ):
                            continue
                        # Now we need to do something
                        # See if we can find the font
                        fname, pos = findfonts.guessFont(style[key])

                        fontList = findfonts.autoEmbed(style[key])
                        if style[key] not in embedded_fontnames and fontList:
                            embedded_fontnames.append(style[key])

                        if not fontList:
                            if (fname, pos) in embedded_fontnames:
                                fontList = None
                            else:
                                fontList = findfonts.autoEmbed(fname)
                            if fontList:
                                embedded_fontnames.append((fname, pos))

                        if fontList:
                            self.embedded += fontList
                            # Maybe the font we got is not called
                            # the same as the one we gave so check that out
                            suff = ["", "-Bold", "-Oblique", "-BoldOblique"]
                            if not fontList[0].startswith(style[key]):
                                # We need to create font aliases, and use them
                                basefname = style[key].split('-')[0]
                                for fname, aliasname in zip(
                                    fontList,
                                    [basefname + suffix for suffix in suff],
                                ):
                                    self.fontsAlias[aliasname] = fname
                                style[key] = self.fontsAlias[basefname + suff[pos]]
                        else:
                            log.error(
                                "Unknown font: \"%s\"," "replacing with Helvetica",
                                style[key],
                            )
                            style[key] = "Helvetica"

        # Get styles from all stylesheets in order
        self.stylesheet = {}
        self.styles = []
        self.linkColor = 'navy'
        # FIXME: linkColor should probably not be a global
        #        style, and tocColor should probably not
        #        be a special case, but for now I'm going
        #        with the flow...
        self.tocColor = None
        for data, ssname in ssdata:
            self.linkColor = data.get('linkColor') or self.linkColor
            self.tocColor = data.get('tocColor') or self.tocColor
            for [skey, style] in self.stylepairs(data):
                sdict = {}
                # FIXME: this is done completely backwards
                for key in style:
                    # Handle color references by name
                    if key == 'color' or key.endswith('Color') and style[key]:
                        style[key] = formatColor(style[key])
                    elif key == 'commands':
                        style[key] = validateCommands(style[key])

                    # Handle alignment constants
                    elif key == 'alignment':
                        style[key] = dict(
                            TA_LEFT=0,
                            LEFT=0,
                            TA_CENTER=1,
                            CENTER=1,
                            TA_CENTRE=1,
                            CENTRE=1,
                            TA_RIGHT=2,
                            RIGHT=2,
                            TA_JUSTIFY=4,
                            JUSTIFY=4,
                            DECIMAL=8,
                        )[style[key].upper()]

                    elif key == 'language':
                        if not style[key] in self.languages:
                            self.languages.append(style[key])

                    sdict[key] = style[key]
                    sdict['name'] = skey
                # If the style already exists, update it
                if skey in self.stylesheet:
                    self.stylesheet[skey].update(sdict)
                else:  # New style
                    self.stylesheet[skey] = sdict
                    self.styles.append(sdict)

        # If the stylesheet has a style name docutils won't reach
        # make a copy with a sanitized name.
        # This may make name collisions possible but that should be
        # rare (who would have custom_name and custom-name in the
        # same stylesheet? ;-)
        # Issue 339

        styles2 = []
        for s in self.styles:
            if not re.match("^[a-z](-?[a-z0-9]+)*$", s['name']):
                s2 = copy(s)
                s2['name'] = docutils.nodes.make_id(s['name'])
                log.warning(
                    '%s is an invalid docutils class name, adding alias %s'
                    % (s['name'], s2['name'])
                )
                styles2.append(s2)
        self.styles.extend(styles2)

        # And create  reportlabs stylesheet
        self.StyleSheet = StyleSheet1()

        dirty = True
        while dirty:
            dirty = False
            for s in self.styles:
                if s['name'] in self.StyleSheet:
                    continue
                try:
                    if 'parent' in s:
                        if s['parent'] is None:
                            if s['name'] != 'base':
                                s['parent'] = self.StyleSheet['base']
                            else:
                                del s['parent']
                        else:
                            s['parent'] = self.StyleSheet[s['parent']]
                    else:
                        if s['name'] != 'base':
                            s['parent'] = self.StyleSheet['base']
                except KeyError:
                    dirty = True
                    continue

                # If the style has no bulletFontName but it has a fontName, set it
                if ('bulletFontName' not in s) and ('fontName' in s):
                    s['bulletFontName'] = s['fontName']

                hasFS = True
                # Adjust fontsize units
                if 'fontSize' not in s:
                    s['fontSize'] = s['parent'].fontSize
                    s['trueFontSize'] = None
                    hasFS = False
                elif 'parent' in s:
                    # This means you can set the fontSize to
                    # "2cm" or to "150%" which will be calculated
                    # relative to the parent style
                    s['fontSize'] = self.adjustUnits(
                        s['fontSize'], s['parent'].fontSize
                    )
                    s['trueFontSize'] = s['fontSize']
                else:
                    # If s has no parent, it's base, which has
                    # an explicit point size by default and %
                    # makes no sense, but guess it as % of 10pt
                    s['fontSize'] = self.adjustUnits(s['fontSize'], 10)

                # If the leading is not set, but the size is, set it
                if 'leading' not in s and hasFS:
                    s['leading'] = 1.2 * s['fontSize']

                # If the bullet font size is not set, set it as fontSize
                if ('bulletFontSize' not in s) and ('fontSize' in s):
                    s['bulletFontSize'] = s['fontSize']

                if 'spaceBefore' in s:
                    if isinstance(s['spaceBefore'], str) and s[
                        'spaceBefore'
                    ].startswith('-'):
                        log.warning('A negative spaceBefore is the same as 0')
                    s['spaceBefore'] = self.adjustUnits(s['spaceBefore'])
                if 'spaceAfter' in s:
                    if isinstance(s['spaceAfter'], str) and s['spaceAfter'].startswith(
                        '-'
                    ):
                        log.warning('A negative spaceAfter is the same as 0')
                    s['spaceAfter'] = self.adjustUnits(s['spaceAfter'])

                self.StyleSheet.add(ParagraphStyle(**s))

        self.emsize = self['base'].fontSize
        # Set the basefont, for Issue 65
        reportlab.rl_config.canvas_basefontname = self['base'].fontName
        # Set a default font for table cell styles (Issue 65)
        reportlab.platypus.tables.CellStyle.fontname = self['base'].fontName

    def __getitem__(self, key):

        # This 'normalizes' the key.
        # For example, if the key is todo_node (like sphinx uses), it will be
        # converted to 'todo-node' which is a valid docutils class name.

        if not re.match("^[a-z](-?[a-z0-9]+)*$", key):
            key = docutils.nodes.make_id(key)

        if key in self.StyleSheet:
            return self.StyleSheet[key]
        else:
            # Sphinx 4 doesn't prepend "pygments-" to its style names when using pygments,
            # so look to see if we have a pygemts style for this style name
            pygments_key = "pygments-" + key
            if pygments_key in self.StyleSheet:
                return self.StyleSheet[pygments_key]
            else:
                if key.startswith('pygments'):
                    if not self.suppress_undefined_style_warning:
                        log.info(
                            "Using undefined style '%s'"
                            ", aliased to style 'code'." % key
                        )
                    newst = copy(self.StyleSheet['code'])
                else:
                    if not self.suppress_undefined_style_warning:
                        log.warning(
                            "Using undefined style '%s'"
                            ", aliased to style 'normal'." % key
                        )
                    newst = copy(self.StyleSheet['normal'])
                newst.name = key
                self.StyleSheet.add(newst)
                return newst

    def readSheets(self, flist):
        """Read in the stylesheets.  Return a list of
        (sheetdata, sheetname) tuples.

        Orders included sheets in front
        of including sheets.
        """

        # Process from end of flist
        flist.reverse()
        # Keep previously seen sheets in sheetdict
        sheetdict = {}
        result = []

        while flist:
            ssname = flist.pop()
            data = sheetdict.get(ssname)
            if data is None:
                data = self.readStyle(ssname)
                if data is None:
                    continue
                sheetdict[ssname] = data
                if 'options' in data and 'stylesheets' in data['options']:
                    flist.append(ssname)
                    newsheets = list(data['options']['stylesheets'])
                    newsheets.reverse()
                    flist.extend(newsheets)
                    continue
            result.append((data, ssname))
        return result

    def readStyle(self, ssname):
        # If callables are used, they should probably be subclassed
        # strings, or something else that will print nicely for errors
        if callable(ssname):
            return ssname()

        fname = self.findStyle(ssname)
        if fname:
            if self.record_dependencies:
                self.record_dependencies.add(fname)

            try:
                # TODO no longer needed when we drop support for rson
                # Is it an older rson/json stylesheet with .style extension?
                root_ext = os.path.splitext(fname)
                if root_ext[1] == ".style":
                    log.warning(
                        'Stylesheet "%s" in outdated format, recommend converting to YAML'
                        % (fname)
                    )
                    return rson_loads(open(fname).read())
                # Otherwise assume yaml/yml
                return yaml.safe_load(open(fname).read())
            except ValueError as e:  # Error parsing the JSON data
                log.critical('Error parsing stylesheet "%s": %s' % (fname, str(e)))
            except IOError as e:  # Error opening the ssheet
                log.critical('Error opening stylesheet "%s": %s' % (fname, str(e)))

    def findStyle(self, fn):
        """Find the absolute file name for a given style filename.

        Given a style filename, searches for it in StyleSearchPath
        and returns the real file name.

        """

        def innerFind(path, fn):
            if os.path.isabs(fn):
                if os.path.isfile(fn):
                    return fn
            else:
                for D in path:
                    tfn = os.path.join(D, fn)
                    if os.path.isfile(tfn):
                        return tfn
            return None

        for ext in ['', '.yaml', '.yml', '.style', '.json']:
            result = innerFind(self.StyleSearchPath, fn + ext)
            if result:
                break
        if result is None:
            log.warning("Can't find stylesheet %s" % fn)
        return result

    def findFont(self, fn):
        """Find the absolute font name for a given font filename.

        Given a font filename, searches for it in FontSearchPath
        and returns the real file name.

        """
        if not os.path.isabs(fn):
            for D in self.FontSearchPath:
                tfn = os.path.join(D, fn)
                if os.path.isfile(tfn):
                    return str(tfn)
        return str(fn)

    def styleForNode(self, node):
        """Return the right default style for any kind of node.

        That usually means "bodytext", but for sidebars, for
        example, it's sidebar.

        """
        n = docutils.nodes
        styles = {
            n.sidebar: 'sidebar',
            n.figure: 'figure',
            n.tgroup: 'table',
            n.table: 'table',
            n.Admonition: 'admonition',
        }

        return self[styles.get(node.__class__, 'bodytext')]

    def tstyleHead(self, rows=1):
        """Return a table style spec for a table header of `rows`.

        The style will be based on the table-heading style from the stylesheet.

        """
        # This alignment thing is exactly backwards from
        # the alignment for paragraphstyles
        alignment = {0: 'LEFT', 1: 'CENTER', 2: 'RIGHT', 4: 'JUSTIFY', 8: 'DECIMAL'}[
            self['table-heading'].alignment
        ]
        return [
            (
                'BACKGROUND',
                (0, 0),
                (-1, rows - 1),
                self['table-heading'].backColor,
            ),
            ('ALIGN', (0, 0), (-1, rows - 1), alignment),
            (
                'TEXTCOLOR',
                (0, 0),
                (-1, rows - 1),
                self['table-heading'].textColor,
            ),
            (
                'FONT',
                (0, 0),
                (-1, rows - 1),
                self['table-heading'].fontName,
                self['table-heading'].fontSize,
                self['table-heading'].leading,
            ),
            ('VALIGN', (0, 0), (-1, rows - 1), self['table-heading'].valign),
        ]

    def adjustUnits(self, v, total=None, default_unit='pt'):
        if total is None:
            total = self.tw
        return adjustUnits(v, total, self.def_dpi, default_unit, emsize=self.emsize)

    def combinedStyle(self, styles):
        """Given a list of style names, it merges them (the existing ones)
        and returns a new style.

        The styles that don't exist are silently ignored.

        For example, if called with styles=['style1','style2'] the returned
        style will be called 'merged_style1_style2'.

        The styles that are *later* in the list will have priority.
        """

        validst = [x for x in styles if x in self.StyleSheet]
        newname = '_'.join(['merged'] + validst)
        validst = [self[x] for x in validst]
        newst = copy(validst[0])

        for st in validst[1:]:
            newst.__dict__.update(st.__dict__)

        newst.name = newname
        return newst


def adjustUnits(v, total=None, dpi=300, default_unit='pt', emsize=10):
    """Takes something like 2cm and returns 2*cm.

    If you use % as a unit, it returns the percentage of "total".

    If total is not given, returns a percentage of the page width.
    However, if you get to that stage, you are doing it wrong.

    Example::

            >>> adjustUnits('50%',200)
            100

    """

    if v is None or v == "":
        return None

    v = str(v)
    length = re.split(r'(-?[0-9.]+)', v)
    n = length[1]
    u = default_unit
    if len(length) == 3 and length[2]:
        u = length[2]
    if u in units.__dict__:
        return float(n) * units.__dict__[u]
    else:
        if u == '%':
            return float(n) * total / 100
        elif u == 'px':
            return float(n) * units.inch / dpi
        elif u == 'pt':
            return float(n)
        elif u == 'in':
            return float(n) * units.inch
        elif u == 'em':
            return float(n) * emsize
        elif u == 'ex':
            return float(n) * emsize / 2
        elif u == 'pc':  # picas!
            return float(n) * 12
        log.error('Unknown unit "%s"' % u)
    return float(n)


def formatColor(value, numeric=True):
    """Convert a color like "gray" or "0xf" or "ffff"
    to something ReportLab will like."""
    if value in colors.__dict__:
        return colors.__dict__[value]
    else:  # Hopefully, a hex color:
        c = value.strip()
        if c[0] == '#':
            c = c[1:]
        while len(c) < 6:
            c = '0' + c
        if numeric:
            r = int(c[:2], 16) / 255.0
            g = int(c[2:4], 16) / 255.0
            b = int(c[4:6], 16) / 255.0
            if len(c) >= 8:
                alpha = int(c[6:8], 16) / 255.0
                return colors.Color(r, g, b, alpha=alpha)
            return colors.Color(r, g, b)
        else:
            return str("#" + c)


# The values are:
# * Minimum number of arguments
# * Maximum number of arguments
# * Valid types of arguments.
#
# For example, if option FOO takes a list a string and a number,
# but the number is optional:
#
# "FOO":(2,3,"list","string","number")
#
# The reportlab command could look like
#
# ["FOO",(0,0),(-1,-1),[1,2],"whatever",4]
#
# THe (0,0) (-1,-1) are start and stop and are mandatory.
#
# Possible types of arguments are string, number, color, colorlist


validCommands = {
    # Cell format commands
    "FONT": (1, 3, "string", "number", "number"),
    "FONTNAME": (1, 1, "string"),
    "FACE": (1, 1, "string"),
    "FONTSIZE": (1, 1, "number"),
    "SIZE": (1, 1, "number"),
    "LEADING": (1, 1, "number"),
    "TEXTCOLOR": (1, 1, "color"),
    "ALIGNMENT": (1, 1, "string"),
    "ALIGN": (1, 1, "string"),
    "LEFTPADDING": (1, 1, "number"),
    "RIGHTPADDING": (1, 1, "number"),
    "TOPPADDING": (1, 1, "number"),
    "BOTTOMPADDING": (1, 1, "number"),
    "BACKGROUND": (1, 1, "color"),
    "ROWBACKGROUNDS": (1, 1, "colorlist"),
    "COLBACKGROUNDS": (1, 1, "colorlist"),
    "VALIGN": (1, 1, "string"),
    # Line commands
    "GRID": (2, 2, "number", "color"),
    "BOX": (2, 2, "number", "color"),
    "OUTLINE": (2, 2, "number", "color"),
    "INNERGRID": (2, 2, "number", "color"),
    "LINEBELOW": (2, 2, "number", "color"),
    "LINEABOVE": (2, 2, "number", "color"),
    "LINEBEFORE": (2, 2, "number", "color"),
    "LINEAFTER": (2, 2, "number", "color"),
    # You should NOT have span commands, man!
    # "SPAN":(,,),
}


def validateCommands(commands):
    """Given a list of reportlab's table commands, it fixes some common errors
    and/or removes commands that can't be fixed"""

    fixed = []

    for command in commands:
        command[0] = command[0].upper()
        flag = False
        # See if the command is valid
        if command[0] not in validCommands:
            log.error('Unknown table command %s in stylesheet', command[0])
            continue

        # See if start and stop are the right types
        if not isinstance(command[1], (list, tuple)):
            log.error(
                'Start cell in table command should be list or tuple, got %s [%s]',
                type(command[1]),
                command[1],
            )
            flag = True

        if not isinstance(command[2], (list, tuple)):
            log.error(
                'Stop cell in table command should be list or tuple, got %s [%s]',
                type(command[1]),
                command[1],
            )
            flag = True

        # See if the number of arguments is right
        length = len(command) - 3
        if length > validCommands[command[0]][1]:
            log.error('Too many arguments in table command: %s', command)
            flag = True

        if length < validCommands[command[0]][0]:
            log.error('Too few arguments in table command: %s', command)
            flag = True

        # Validate argument types
        for pos, arg in enumerate(command[3:]):
            typ = validCommands[command[0]][pos + 2]
            if typ == "color":
                # Convert all 'string' colors to numeric
                command[3 + pos] = formatColor(arg)
            elif typ == "colorlist":
                command[3 + pos] = [formatColor(c) for c in arg]
            elif typ == "number":
                pass
            elif typ == "string":
                command[3 + pos] = arg
            else:
                log.error("This should never happen: wrong type %s", typ)

        if not flag:
            fixed.append(command)

    return fixed


class CallableStyleSheet(str):
    """Useful for programmatically generated stylesheets.
    A generated stylesheet is a callable string (name),
    which returns the pre-digested stylesheet data
    when called.
    """

    def __new__(cls, name, value=''):
        self = str.__new__(cls, name)
        self.value = value
        return self

    def __call__(self):
        return rson_loads(self.value)



# -*- coding: utf-8 -*-
import codecs
import tempfile

import docutils
from reportlab.platypus.paragraph import Paragraph

import rst2pdf.genelements as genelements
from rst2pdf.flowables import Heading, MyPageBreak
from rst2pdf.image import MyImage


class FancyTitleHandler(genelements.HandleParagraph, docutils.nodes.title):
    """
    This class will handle title nodes.

    It takes a "titletemplate.svg", replaces TITLEGOESHERE with
    the actual title text, and draws that using the FancyHeading flowable
    (see below).

    Since this class is defined in an extension, it
    effectively replaces rst2pdf.genelements.HandleTitle.
    """

    def gather_elements(self, client, node, style):
        # This method is copied from the HandleTitle class
        # in rst2pdf.genelements.

        # Special cases: (Not sure this is right ;-)
        if isinstance(node.parent, docutils.nodes.document):
            # node.elements = [Paragraph(client.gen_pdftext(node),
            # client.styles['title'])]
            # The visible output is now done by the cover template
            node.elements = []
            client.doc_title = node.rawsource
            client.doc_title_clean = node.astext().strip()
        elif isinstance(node.parent, docutils.nodes.topic):
            node.elements = [
                Paragraph(client.gen_pdftext(node), client.styles['topic-title'])
            ]
        elif isinstance(node.parent, docutils.nodes.Admonition):
            node.elements = [
                Paragraph(client.gen_pdftext(node), client.styles['admonition-title'])
            ]
        elif isinstance(node.parent, docutils.nodes.table):
            node.elements = [
                Paragraph(client.gen_pdftext(node), client.styles['table-title'])
            ]
        elif isinstance(node.parent, docutils.nodes.sidebar):
            node.elements = [
                Paragraph(client.gen_pdftext(node), client.styles['sidebar-title'])
            ]
        else:
            # Section/Subsection/etc.
            text = client.gen_pdftext(node)
            fch = node.children[0]
            if isinstance(fch, docutils.nodes.generated) and fch['classes'] == [
                'sectnum'
            ]:
                snum = fch.astext()
            else:
                snum = None
            maxdepth = 6

            # The parent ID is the refid + an ID to make it unique for Sphinx
            parent_id = (node.parent.get('ids', [None]) or [None])[0] + u'-%s' % id(
                node
            )
            if client.depth > 1:
                node.elements = [
                    Heading(
                        text,
                        client.styles['heading%d' % min(client.depth, maxdepth)],
                        level=client.depth - 1,
                        parent_id=parent_id,
                        node=node,
                    )
                ]
            else:  # This is an important title, do our magic ;-)
                # Hack the title template SVG
                tfile = codecs.open('titletemplate.svg', 'r', 'utf-8')
                tdata = tfile.read()
                tfile.close()
                tfile = tempfile.NamedTemporaryFile(
                    dir='.', delete=False, suffix='.svg'
                )
                tfname = tfile.name
                tfile.write(tdata.replace('TITLEGOESHERE', text).encode('utf-8'))
                tfile.close()

                # Now tfname contains a SVG with the right title.
                # Make rst2pdf delete it later.
                client.to_unlink.append(tfname)

                e = FancyHeading(
                    tfname,
                    width=700,
                    height=100,
                    client=client,
                    snum=snum,
                    parent_id=parent_id,
                    text=text,
                    hstyle=client.styles['heading%d' % min(client.depth, maxdepth)],
                )

                node.elements = [e]

            if client.depth <= client.breaklevel:
                node.elements.insert(0, MyPageBreak(breakTo=client.breakside))
        return node.elements


class FancyHeading(MyImage, Heading):
    """This is a cross between the Heading flowable, that adds outline
    entries so you have a PDF TOC, and MyImage, that draws images"""

    def __init__(self, *args, **kwargs):
        # The inicialization is taken from rst2pdf.flowables.Heading
        hstyle = kwargs.pop('hstyle')
        level = 0
        text = kwargs.pop('text')
        self.snum = kwargs.pop('snum')
        self.parent_id = kwargs.pop('parent_id')
        # self.stext =
        Heading.__init__(self, text, hstyle, level=level, parent_id=self.parent_id)
        # Cleanup title text
        # self.stext = re.sub(r'<[^>]*?>', '', unescape(self.stext))
        # self.stext = self.stext.strip()

        # Stuff needed for the outline entry
        MyImage.__init__(self, *args, **kwargs)

    def drawOn(self, canv, x, y, _sW):
        # Add outline entry. This is copied from rst2pdf.flowables.heading
        canv.bookmarkHorizontal(self.parent_id, 0, y + self.image.height)

        if canv.firstSect:
            canv.sectName = self.stext
            canv.firstSect = False
            if self.snum is not None:
                canv.sectNum = self.snum
            else:
                canv.sectNum = ""

        canv.addOutlineEntry(self.stext, self.parent_id, int(self.level), False)

        # And let MyImage do all the drawing
        MyImage.drawOn(self, canv, x, y, _sW)
